# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from datetime import datetime

from airflow.providers.amazon.aws.operators.comprehend import ComprehendStartPiiEntitiesDetectionJobOperator
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.comprehend import (
    ComprehendStartPiiEntitiesDetectionJobCompletedSensor,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task_group
else:
    # Airflow 2 path
    from airflow.decorators import task_group  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import SystemTestContextBuilder

ROLE_ARN_KEY = "ROLE_ARN"
sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

DAG_ID = "example_comprehend"
INPUT_S3_KEY_START_PII_ENTITIES_DETECTION_JOB = "start-pii-entities-detection-job/sample_data.txt"

SAMPLE_DATA = {
    "username": "bob1234",
    "name": "Bob",
    "sex": "M",
    "address": "1773 Raymond Ville Suite 682",
    "mail": "test@hotmail.com",
}


@task_group
def pii_entities_detection_job_workflow():
    # [START howto_operator_start_pii_entities_detection_job]
    start_pii_entities_detection_job = ComprehendStartPiiEntitiesDetectionJobOperator(
        task_id="start_pii_entities_detection_job",
        input_data_config=input_data_configurations,
        output_data_config=output_data_configurations,
        mode="ONLY_REDACTION",
        data_access_role_arn=test_context[ROLE_ARN_KEY],
        language_code="en",
        start_pii_entities_kwargs=pii_entities_kwargs,
    )
    # [END howto_operator_start_pii_entities_detection_job]
    start_pii_entities_detection_job.wait_for_completion = False

    # [START howto_sensor_start_pii_entities_detection_job]
    await_start_pii_entities_detection_job = ComprehendStartPiiEntitiesDetectionJobCompletedSensor(
        task_id="await_start_pii_entities_detection_job", job_id=start_pii_entities_detection_job.output
    )
    # [END howto_sensor_start_pii_entities_detection_job]

    chain(start_pii_entities_detection_job, await_start_pii_entities_detection_job)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]
    bucket_name = f"{env_id}-comprehend"
    input_data_configurations = {
        "S3Uri": f"s3://{bucket_name}/{INPUT_S3_KEY_START_PII_ENTITIES_DETECTION_JOB}",
        "InputFormat": "ONE_DOC_PER_LINE",
    }
    output_data_configurations = {"S3Uri": f"s3://{bucket_name}/redacted_output/"}
    pii_entities_kwargs = {
        "RedactionConfig": {
            "PiiEntityTypes": ["NAME", "ADDRESS"],
            "MaskMode": "REPLACE_WITH_PII_ENTITY_TYPE",
        }
    }

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    upload_sample_data = S3CreateObjectOperator(
        task_id="upload_sample_data",
        s3_bucket=bucket_name,
        s3_key=INPUT_S3_KEY_START_PII_ENTITIES_DETECTION_JOB,
        data=json.dumps(SAMPLE_DATA),
    )

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        upload_sample_data,
        # TEST BODY
        pii_entities_detection_job_workflow(),
        # TEST TEARDOWN
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime, timezone
from unittest import mock

import pytest

from airflow.providers.amazon.aws.executors.utils.exponential_backoff_retry import (
    calculate_next_attempt_delay,
    exponential_backoff_retry,
)


class TestExponentialBackoffRetry:
    def test_exponential_backoff_retry_base_case(self, time_machine):
        time_machine.move_to(datetime(2023, 1, 1, 12, 0, 5))
        mock_callable_function = mock.Mock()
        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=0,
            callable_function=mock_callable_function,
        )
        mock_callable_function.assert_called_once()

    @pytest.mark.parametrize(
        ("attempt_number", "utcnow_value", "expected_calls"),
        [
            (
                0,
                datetime(2023, 1, 1, 12, 0, 2),
                1,
            ),
            (
                1,
                datetime(2023, 1, 1, 12, 0, 3),
                0,
            ),  # delay is 4 seconds; no call made
            (
                1,
                datetime(2023, 1, 1, 12, 0, 4),
                1,
            ),
            (
                2,
                datetime(2023, 1, 1, 12, 0, 15),
                0,
            ),  # delay is 16 seconds; no call made
            (
                2,
                datetime(2023, 1, 1, 12, 0, 16),
                1,
            ),
            (
                3,
                datetime(2023, 1, 1, 12, 1, 3),
                0,
            ),  # delay is 64 seconds; no call made
            (
                3,
                datetime(2023, 1, 1, 12, 1, 4),
                1,
            ),
            (
                4,
                datetime(2023, 1, 1, 12, 1, 59),
                0,
            ),  # delay is 120 seconds; no call made
            (
                4,
                datetime(2023, 1, 1, 12, 2, 0),
                1,
            ),
            (
                5,
                datetime(2023, 1, 1, 12, 1, 59),
                0,
            ),  # delay is 120 seconds; no call made
            (
                5,
                datetime(2023, 1, 1, 12, 2, 0),
                1,
            ),
            (
                99,
                datetime(2023, 1, 1, 12, 1, 59),
                0,
            ),  # delay is 120 seconds; no call made
            (
                99,
                datetime(2023, 1, 1, 12, 2, 0),
                1,
            ),
        ],
    )
    def test_exponential_backoff_retry_parameterized(
        self, attempt_number, utcnow_value, expected_calls, time_machine
    ):
        time_machine.move_to(utcnow_value)
        mock_callable_function = mock.Mock()
        mock_callable_function.__name__ = "test_callable_function"
        mock_callable_function.side_effect = Exception()

        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=attempt_number,
            callable_function=mock_callable_function,
        )
        assert mock_callable_function.call_count == expected_calls

    def test_exponential_backoff_retry_fail_success(self, time_machine, caplog):
        mock_callable_function = mock.Mock()
        mock_callable_function.__name__ = "test_callable_function"
        mock_callable_function.side_effect = [Exception(), True]
        time_machine.move_to(datetime(2023, 1, 1, 12, 0, 2))
        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=0,
            callable_function=mock_callable_function,
        )
        mock_callable_function.assert_called_once()
        assert any("Error calling" in log for log in caplog.messages)
        caplog.clear()  # clear messages so that we have clean logs for the next call

        time_machine.move_to(datetime(2023, 1, 1, 12, 0, 6))
        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=1,
            callable_function=mock_callable_function,
        )
        assert all("Error calling" not in log for log in caplog.messages)

    def test_exponential_backoff_retry_max_delay(self, time_machine):
        mock_callable_function = mock.Mock()
        mock_callable_function.__name__ = "test_callable_function"
        mock_callable_function.return_value = Exception()
        time_machine.move_to(datetime(2023, 1, 1, 12, 4, 15))
        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=4,
            callable_function=mock_callable_function,
            max_delay=60 * 5,
        )
        mock_callable_function.assert_not_called()  # delay is 256 seconds; no calls made
        time_machine.move_to(datetime(2023, 1, 1, 12, 4, 16))
        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=4,
            callable_function=mock_callable_function,
            max_delay=60 * 5,
        )
        mock_callable_function.assert_called_once()

        time_machine.move_to(datetime(2023, 1, 1, 12, 5, 0))
        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=5,
            callable_function=mock_callable_function,
            max_delay=60 * 5,
        )
        # delay should be 4^5=1024 seconds, but max_delay is 60*5=300 seconds
        assert mock_callable_function.call_count == 2

    def test_exponential_backoff_retry_max_attempts(self, time_machine, caplog):
        mock_callable_function = mock.Mock()
        mock_callable_function.__name__ = "test_callable_function"
        mock_callable_function.return_value = Exception()
        time_machine.move_to(datetime(2023, 1, 1, 12, 55, 0))
        for i in range(10):
            exponential_backoff_retry(
                last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
                attempts_since_last_successful=i,
                callable_function=mock_callable_function,
                max_attempts=3,
            )
        assert any("Max attempts reached." in log for log in caplog.messages)
        assert mock_callable_function.call_count == 3

    @pytest.mark.parametrize(
        ("attempt_number", "utcnow_value", "expected_calls"),
        [
            (
                0,
                datetime(2023, 1, 1, 12, 0, 2),
                1,
            ),
            (
                1,
                datetime(2023, 1, 1, 12, 0, 2),
                0,
            ),  # delay is 3 seconds; no call made
            (
                1,
                datetime(2023, 1, 1, 12, 0, 3),
                1,
            ),
            (
                2,
                datetime(2023, 1, 1, 12, 0, 8),
                0,
            ),  # delay is 9 seconds; no call made
            (
                2,
                datetime(2023, 1, 1, 12, 0, 9),
                1,
            ),
            (
                3,
                datetime(2023, 1, 1, 12, 0, 26),
                0,
            ),  # delay is 27 seconds; no call made
            (
                3,
                datetime(2023, 1, 1, 12, 0, 27),
                1,
            ),
            (
                4,
                datetime(2023, 1, 1, 12, 1, 20),
                0,
            ),  # delay is 81 seconds; no call made
            (
                4,
                datetime(2023, 1, 1, 12, 1, 21),
                1,
            ),
            (
                5,
                datetime(2023, 1, 1, 12, 1, 59),
                0,
            ),  # delay is 120 seconds; no call made
            (
                5,
                datetime(2023, 1, 1, 12, 2, 0),
                1,
            ),
            (
                99,
                datetime(2023, 1, 1, 12, 1, 59),
                0,
            ),  # delay is 120 seconds; no call made
            (
                99,
                datetime(2023, 1, 1, 12, 2, 0),
                1,
            ),
        ],
    )
    def test_exponential_backoff_retry_exponent_base_parameterized(
        self, time_machine, attempt_number, utcnow_value, expected_calls
    ):
        mock_callable_function = mock.Mock()
        mock_callable_function.__name__ = "test_callable_function"
        mock_callable_function.side_effect = Exception()
        time_machine.move_to(utcnow_value)

        exponential_backoff_retry(
            last_attempt_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            attempts_since_last_successful=attempt_number,
            callable_function=mock_callable_function,
            exponent_base=3,
        )
        assert mock_callable_function.call_count == expected_calls

    def test_calculate_next_attempt_delay(self):
        exponent_base: int = 4
        num_loops: int = 3
        # Setting max_delay this way means there will be three loops will run to test:
        # one will return a value under max_delay, one equal to max_delay, and one over.
        max_delay: int = exponent_base**num_loops - 1

        for attempt_number in range(1, num_loops):
            returned_delay = calculate_next_attempt_delay(attempt_number, max_delay, exponent_base).seconds

            if (expected_delay := exponent_base**attempt_number) <= max_delay:
                assert returned_delay == expected_delay
            else:
                assert returned_delay == max_delay



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.glue_crawler import GlueCrawlerHook
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.triggers.glue_crawler import GlueCrawlerCompleteTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class GlueCrawlerOperator(AwsBaseOperator[GlueCrawlerHook]):
    """
    Creates, updates and triggers an AWS Glue Crawler.

    AWS Glue Crawler is a serverless service that manages a catalog of
    metadata tables that contain the inferred schema, format and data
    types of data stores within the AWS cloud.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GlueCrawlerOperator`

    :param config: Configurations for the AWS Glue crawler
    :param poll_interval: Time (in seconds) to wait between two consecutive calls to check crawler status
    :param wait_for_completion: Whether to wait for crawl execution completion. (default: True)
    :param deferrable: If True, the operator will wait asynchronously for the crawl to complete.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = GlueCrawlerHook

    template_fields: Sequence[str] = aws_template_fields(
        "config",
    )
    ui_color = "#ededed"

    def __init__(
        self,
        config,
        poll_interval: int = 5,
        wait_for_completion: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.poll_interval = poll_interval
        self.wait_for_completion = wait_for_completion
        self.deferrable = deferrable
        self.config = config

    def execute(self, context: Context) -> str:
        """
        Execute AWS Glue Crawler from Airflow.

        :return: the name of the current glue crawler.
        """
        crawler_name = self.config["Name"]
        if self.hook.has_crawler(crawler_name):
            self.hook.update_crawler(**self.config)
        else:
            self.hook.create_crawler(**self.config)

        self.log.info("Triggering AWS Glue Crawler")
        self.hook.start_crawler(crawler_name)
        if self.deferrable:
            self.defer(
                trigger=GlueCrawlerCompleteTrigger(
                    crawler_name=crawler_name,
                    waiter_delay=self.poll_interval,
                    aws_conn_id=self.aws_conn_id,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
            )
        elif self.wait_for_completion:
            self.log.info("Waiting for AWS Glue Crawler")
            self.hook.wait_for_crawler_completion(crawler_name=crawler_name, poll_interval=self.poll_interval)

        return crawler_name

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> str:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(f"Error in glue crawl: {validated_event}")
        return self.config["Name"]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator, Generator
from typing import Any, TypedDict

from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.utils.helpers import prune_dict

# Guidance received from the AWS team regarding the correct way to check for the end of a stream is that the
# value of the nextForwardToken is the same in subsequent calls.
# The issue with this approach is, it can take a huge amount of time (e.g. 20 seconds) to retrieve logs using
# this approach. As an intermediate solution, we decided to stop fetching logs if 3 consecutive responses
# are empty.
# See PR https://github.com/apache/airflow/pull/20814
NUM_CONSECUTIVE_EMPTY_RESPONSE_EXIT_THRESHOLD = 3


class CloudWatchLogEvent(TypedDict):
    """TypedDict for CloudWatch Log Event."""

    timestamp: int
    message: str
    ingestionTime: int


class AwsLogsHook(AwsBaseHook):
    """
    Interact with Amazon CloudWatch Logs.

    Provide thin wrapper around :external+boto3:py:class:`boto3.client("logs") <CloudWatchLogs.Client>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs["client_type"] = "logs"
        super().__init__(*args, **kwargs)

    class ContinuationToken:
        """Just a wrapper around a str token to allow updating it from the caller."""

        def __init__(self):
            self.value: str | None = None

    def get_log_events(
        self,
        log_group: str,
        log_stream_name: str,
        start_time: int = 0,
        skip: int = 0,
        start_from_head: bool | None = None,
        continuation_token: ContinuationToken | None = None,
        end_time: int | None = None,
    ) -> Generator[CloudWatchLogEvent, None, None]:
        """
        Return a generator for log items in a single stream; yields all items available at the current moment.

        .. seealso::
            - :external+boto3:py:meth:`CloudWatchLogs.Client.get_log_events`

        :param log_group: The name of the log group.
        :param log_stream_name: The name of the specific stream.
        :param start_time: The timestamp value in ms to start reading the logs from (default: 0).
        :param skip: The number of log entries to skip at the start (default: 0).
            This is for when there are multiple entries at the same timestamp.
        :param continuation_token: a token indicating where to read logs from.
            Will be updated as this method reads new logs, to be reused in subsequent calls.
        :param end_time: The timestamp value in ms to stop reading the logs from (default: None).
            If None is provided, reads it until the end of the log stream
        :return: | A CloudWatch log event with the following key-value pairs:
                 |   'timestamp' (int): The time in milliseconds of the event.
                 |   'message' (str): The log event data.
                 |   'ingestionTime' (int): The time in milliseconds the event was ingested.
        """
        if continuation_token is None:
            continuation_token = AwsLogsHook.ContinuationToken()

        num_consecutive_empty_response = 0
        while True:
            if continuation_token.value is not None:
                token_arg: dict[str, str] = {"nextToken": continuation_token.value}
            else:
                token_arg = {}

            response = self.conn.get_log_events(
                **prune_dict(
                    {
                        "logGroupName": log_group,
                        "logStreamName": log_stream_name,
                        "startTime": start_time,
                        "endTime": end_time,
                        "startFromHead": True,
                        **token_arg,
                    }
                )
            )

            events = response["events"]
            event_count = len(events)

            if event_count > skip:
                events = events[skip:]
                skip = 0
            else:
                skip -= event_count
                events = []

            yield from events

            if continuation_token.value == response["nextForwardToken"]:
                return

            if not event_count:
                num_consecutive_empty_response += 1
                if num_consecutive_empty_response >= NUM_CONSECUTIVE_EMPTY_RESPONSE_EXIT_THRESHOLD:
                    # Exit if there are more than NUM_CONSECUTIVE_EMPTY_RESPONSE_EXIT_THRESHOLD consecutive
                    # empty responses
                    return
            else:
                num_consecutive_empty_response = 0

            continuation_token.value = response["nextForwardToken"]

    async def describe_log_streams_async(
        self, log_group: str, stream_prefix: str, order_by: str, count: int
    ) -> dict[str, Any] | None:
        """
        Async function to get the list of log streams for the specified log group.

        You can list all the log streams or filter the results by prefix. You can also control
        how the results are ordered.

        :param log_group: The name of the log group.
        :param stream_prefix: The prefix to match.
        :param order_by: If the value is LogStreamName , the results are ordered by log stream name.
         If the value is LastEventTime , the results are ordered by the event time. The default value is LogStreamName.
        :param count: The maximum number of items returned
        """
        async with await self.get_async_conn() as client:
            try:
                response: dict[str, Any] = await client.describe_log_streams(
                    logGroupName=log_group,
                    logStreamNamePrefix=stream_prefix,
                    orderBy=order_by,
                    limit=count,
                )
                return response
            except ClientError as error:
                # On the very first training job run on an account, there's no log group until
                # the container starts logging, so ignore any errors thrown about that
                if error.response["Error"]["Code"] == "ResourceNotFoundException":
                    return None
                raise error

    async def get_log_events_async(
        self,
        log_group: str,
        log_stream_name: str,
        start_time: int = 0,
        skip: int = 0,
        start_from_head: bool = True,
    ) -> AsyncGenerator[Any, dict[str, Any]]:
        """
        Yield all the available items in a single log stream.

        :param log_group: The name of the log group.
        :param log_stream_name: The name of the specific stream.
        :param start_time: The time stamp value to start reading the logs from (default: 0).
        :param skip: The number of log entries to skip at the start (default: 0).
            This is for when there are multiple entries at the same timestamp.
        :param start_from_head: whether to start from the beginning (True) of the log or
            at the end of the log (False).
        """
        next_token = None
        while True:
            if next_token is not None:
                token_arg: dict[str, str] = {"nextToken": next_token}
            else:
                token_arg = {}

            async with await self.get_async_conn() as client:
                response = await client.get_log_events(
                    logGroupName=log_group,
                    logStreamName=log_stream_name,
                    startTime=start_time,
                    startFromHead=start_from_head,
                    **token_arg,
                )

                events = response["events"]
                event_count = len(events)

                if event_count > skip:
                    events = events[skip:]
                    skip = 0
                else:
                    skip -= event_count
                    events = []

                for event in events:
                    await asyncio.sleep(1)
                    yield event

                if next_token != response["nextForwardToken"]:
                    next_token = response["nextForwardToken"]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

import pytest

from airflow.providers.amazon.aws.hooks.mwaa import MwaaHook
from airflow.providers.amazon.aws.operators.mwaa import MwaaTriggerDagRunOperator

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

OP_KWARGS = {
    "task_id": "test_task",
    "env_name": "test_env",
    "trigger_dag_id": "test_dag_id",
    "trigger_run_id": "test_run_id",
    "logical_date": "2025-01-01T00:00:01Z",
    "data_interval_start": "2025-01-02T00:00:01Z",
    "data_interval_end": "2025-01-03T00:00:01Z",
    "conf": {"key": "value"},
    "note": "test note",
    "wait_for_completion": False,
    "waiter_delay": 5,
    "waiter_max_attempts": 20,
    "deferrable": False,
}
HOOK_RETURN_VALUE = {
    "ResponseMetadata": {},
    "RestApiStatusCode": 200,
    "RestApiResponse": {
        "dag_run_id": "manual__2025-02-08T00:33:09.457198+00:00",
        "other_key": "value",
    },
}


class TestMwaaTriggerDagRunOperator:
    def test_init(self):
        op = MwaaTriggerDagRunOperator(**OP_KWARGS)
        assert op.env_name == OP_KWARGS["env_name"]
        assert op.trigger_dag_id == OP_KWARGS["trigger_dag_id"]
        assert op.trigger_run_id == OP_KWARGS["trigger_run_id"]
        assert op.logical_date == OP_KWARGS["logical_date"]
        assert op.data_interval_start == OP_KWARGS["data_interval_start"]
        assert op.data_interval_end == OP_KWARGS["data_interval_end"]
        assert op.conf == OP_KWARGS["conf"]
        assert op.note == OP_KWARGS["note"]
        assert op.wait_for_completion == OP_KWARGS["wait_for_completion"]
        assert op.waiter_delay == OP_KWARGS["waiter_delay"]
        assert op.waiter_max_attempts == OP_KWARGS["waiter_max_attempts"]
        assert op.deferrable == OP_KWARGS["deferrable"]

    @mock.patch.object(MwaaTriggerDagRunOperator, "hook")
    def test_execute(self, mock_hook):
        mock_hook.invoke_rest_api.return_value = HOOK_RETURN_VALUE
        op = MwaaTriggerDagRunOperator(**OP_KWARGS)
        op_ret_val = op.execute({})

        mock_hook.invoke_rest_api.assert_called_once_with(
            env_name=OP_KWARGS["env_name"],
            path=f"/dags/{OP_KWARGS['trigger_dag_id']}/dagRuns",
            method="POST",
            body={
                "dag_run_id": OP_KWARGS["trigger_run_id"],
                "logical_date": OP_KWARGS["logical_date"],
                "data_interval_start": OP_KWARGS["data_interval_start"],
                "data_interval_end": OP_KWARGS["data_interval_end"],
                "conf": OP_KWARGS["conf"],
                "note": OP_KWARGS["note"],
            },
            airflow_version=None,
        )
        assert op_ret_val == HOOK_RETURN_VALUE

    def test_template_fields(self):
        operator = MwaaTriggerDagRunOperator(**OP_KWARGS)
        validate_template_fields(operator)

    @pytest.mark.parametrize(
        ("wait_for_completion", "deferrable"),
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(MwaaHook, "get_waiter")
    @mock.patch.object(MwaaTriggerDagRunOperator, "hook")
    def test_execute_wait_combinations(self, mock_hook, _, wait_for_completion, deferrable):
        kwargs = OP_KWARGS
        kwargs["wait_for_completion"] = wait_for_completion
        kwargs["deferrable"] = deferrable
        op = MwaaTriggerDagRunOperator(**OP_KWARGS)
        mock_hook.invoke_rest_api.return_value = HOOK_RETURN_VALUE
        op.defer = MagicMock()
        response = op.execute({})

        assert response == HOOK_RETURN_VALUE
        assert mock_hook.get_waiter.call_count == wait_for_completion
        assert op.defer.call_count == deferrable



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import boto3
import pytest
from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID

from airflow.providers.amazon.aws.hooks.ecr import EcrHook


@pytest.fixture
def patch_hook(monkeypatch):
    """Patch hook object by dummy boto3 ECR client."""
    ecr_client = boto3.client("ecr")
    monkeypatch.setattr(EcrHook, "conn", ecr_client)


@mock_aws
class TestEcrHook:
    def test_service_type(self):
        """Test expected boto3 client type."""
        assert EcrHook().client_type == "ecr"

    @pytest.mark.parametrize(
        "accounts_ids",
        [
            pytest.param("", id="empty-string"),
            pytest.param(None, id="none"),
            pytest.param([], id="empty-list"),
        ],
    )
    def test_get_temporary_credentials_default_account_id(self, patch_hook, accounts_ids):
        """Test different types of empty account/registry ids."""
        result = EcrHook().get_temporary_credentials(registry_ids=accounts_ids)
        assert len(result) == 1
        assert result[0].username == "AWS"
        assert result[0].registry.startswith(DEFAULT_ACCOUNT_ID)
        assert result[0].password == f"{DEFAULT_ACCOUNT_ID}-auth-token"

    @pytest.mark.parametrize(
        ("accounts_id", "expected_registry"),
        [
            pytest.param(DEFAULT_ACCOUNT_ID, DEFAULT_ACCOUNT_ID, id="moto-default-account"),
            pytest.param("111100002222", "111100002222", id="custom-account-id"),
            pytest.param(["333366669999"], "333366669999", id="custom-account-id-list"),
        ],
    )
    def test_get_temporary_credentials_single_account_id(self, patch_hook, accounts_id, expected_registry):
        """Test different types of single account/registry ids."""
        result = EcrHook().get_temporary_credentials(registry_ids=accounts_id)
        assert len(result) == 1
        assert result[0].username == "AWS"
        assert result[0].registry.startswith(expected_registry)
        assert result[0].password == f"{expected_registry}-auth-token"

    @pytest.mark.parametrize(
        "accounts_ids",
        [
            pytest.param([DEFAULT_ACCOUNT_ID, "111100002222"], id="moto-default-and-custom-account-ids"),
            pytest.param(["999888777666", "333366669999", "777"], id="custom-accounts-ids"),
        ],
    )
    def test_get_temporary_credentials_multiple_account_ids(self, patch_hook, accounts_ids):
        """Test multiple account ids in the single method call."""
        expected_creds = len(accounts_ids)
        result = EcrHook().get_temporary_credentials(registry_ids=accounts_ids)
        assert len(result) == expected_creds
        assert [cr.username for cr in result] == ["AWS"] * expected_creds
        assert all(cr.registry.startswith(accounts_ids[ix]) for ix, cr in enumerate(result))

    @pytest.mark.parametrize(
        "accounts_ids",
        [
            pytest.param(None, id="none"),
            pytest.param("111100002222", id="single-account-id"),
            pytest.param(["999888777666", "333366669999", "777"], id="multiple-account-ids"),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.hooks.ecr.mask_secret")
    def test_get_temporary_credentials_mask_secrets(self, mock_masker, patch_hook, accounts_ids):
        """Test masking passwords."""
        result = EcrHook().get_temporary_credentials(registry_ids=accounts_ids)
        assert mock_masker.call_args_list == [mock.call(cr.password) for cr in result]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from datetime import datetime
from typing import Any
from unittest import mock

import pytest
from dateutil import parser

from airflow.providers.amazon.aws.hooks.dms import DmsHook, DmsTaskWaiterStatus

MOCK_DATA = {
    "replication_task_id": "test_task",
    "source_endpoint_arn": "source-endpoint-arn",
    "target_endpoint_arn": "target-endpoint-arn",
    "replication_instance_arn": "replication-instance-arn",
    "migration_type": "full-load",
    "table_mappings": {
        "rules": [
            {
                "rule-type": "selection",
                "rule-id": "1",
                "rule-name": "1",
                "object-locator": {
                    "schema-name": "test",
                    "table-name": "%",
                },
                "rule-action": "include",
            }
        ]
    },
}
MOCK_TASK_ARN = "task-arn"
MOCK_TASK_RESPONSE_DATA = {
    "ReplicationTaskIdentifier": MOCK_DATA["replication_task_id"],
    "SourceEndpointArn": MOCK_DATA["source_endpoint_arn"],
    "TargetEndpointArn": MOCK_DATA["target_endpoint_arn"],
    "ReplicationInstanceArn": MOCK_DATA["replication_instance_arn"],
    "MigrationType": MOCK_DATA["migration_type"],
    "TableMappings": json.dumps(MOCK_DATA["table_mappings"]),
    "ReplicationTaskArn": MOCK_TASK_ARN,
    "Status": "creating",
}
MOCK_DESCRIBE_RESPONSE: dict[str, Any] = {"ReplicationTasks": [MOCK_TASK_RESPONSE_DATA]}
MOCK_DESCRIBE_RESPONSE_WITH_MARKER: dict[str, Any] = {
    "ReplicationTasks": [MOCK_TASK_RESPONSE_DATA],
    "Marker": "marker",
}
MOCK_CREATE_RESPONSE: dict[str, Any] = {"ReplicationTask": MOCK_TASK_RESPONSE_DATA}
MOCK_START_RESPONSE: dict[str, Any] = {"ReplicationTask": {**MOCK_TASK_RESPONSE_DATA, "Status": "starting"}}
MOCK_STOP_RESPONSE: dict[str, Any] = {"ReplicationTask": {**MOCK_TASK_RESPONSE_DATA, "Status": "stopping"}}
MOCK_DELETE_RESPONSE: dict[str, Any] = {"ReplicationTask": {**MOCK_TASK_RESPONSE_DATA, "Status": "deleting"}}

MOCK_CONFIG_RESPONSE: dict[str, Any] = {
    "Marker": "xxxxx",
    "ReplicationConfigs": [
        {
            "ReplicationConfigIdentifier": "1111",
            "ReplicationConfigArn": "arn:aws:my-arn",
            "SourceEndpointArn": "source-endpoint-arn",
            "TargetEndpointArn": "target-endpoint-arn",
            "ReplicationType": "cdc",
            "ComputeConfig": {
                "AvailabilityZone": "az1",
                "MaxCapacityUnits": 10,
                "MinCapacityUnits": 20,
                "MultiAZ": True,
                "PreferredMaintenanceWindow": "string",
                "ReplicationSubnetGroupId": "string",
                "VpcSecurityGroupIds": [
                    "string",
                ],
            },
            "ReplicationSettings": "string",
            "SupplementalSettings": "string",
            "TableMappings": "string",
            "ReplicationConfigCreateTime": datetime(2015, 1, 1),
            "ReplicationConfigUpdateTime": datetime(2015, 1, 1),
        },
        {
            "ReplicationConfigIdentifier": "2222",
            "ReplicationConfigArn": "arn:aws:my-arn",
            "SourceEndpointArn": "source-endpoint-arn",
            "TargetEndpointArn": "target-endpoint-arn",
            "ReplicationType": "full-load-and-cdc",
            "ComputeConfig": {
                "AvailabilityZone": "string",
                "DnsNameServers": "string",
                "KmsKeyId": "string",
                "MaxCapacityUnits": 1,
                "MinCapacityUnits": 30,
                "MultiAZ": False,
                "PreferredMaintenanceWindow": "string",
                "ReplicationSubnetGroupId": "string",
                "VpcSecurityGroupIds": [
                    "string",
                ],
            },
            "ReplicationSettings": "string",
            "SupplementalSettings": "string",
            "TableMappings": "string",
            "ReplicationConfigCreateTime": datetime(2015, 1, 1),
            "ReplicationConfigUpdateTime": datetime(2015, 2, 1),
        },
    ],
}

MOCK_REPLICATION_CONFIG: dict[str, Any] = {
    "ReplicationConfigIdentifier": "test-config",
    "SourceEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:RZZK4EZW5UANC7Y3P4E776WHBE",
    "TargetEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:GVBUJQXJZASXWHTWCLN2WNT57E",
    "ComputeConfig": {
        "MaxCapacityUnits": 2,
        "MinCapacityUnits": 4,
    },
    "ReplicationType": "full-load",
    "TableMappings": json.dumps(
        {
            "TableMappings": [
                {
                    "Type": "Selection",
                    "RuleId": 123,
                    "RuleName": "test-rule",
                    "SourceSchema": "/",
                    "SourceTable": "/",
                }
            ]
        }
    ),
    "ReplicationSettings": "string",
    "SupplementalSettings": "string",
    "ResourceIdentifier": "string",
}

MOCK_REPLICATION_CONFIG_RESP: dict[str, Any] = {
    "ReplicationConfig": {
        "ReplicationConfigIdentifier": "test-config",
        "ReplicationConfigArn": "arn:aws:dms:us-east-1:123456789012:replication-config/test-config",
        "SourceEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:RZZK4EZW5UANC7Y3P4E776WHBE",
        "TargetEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:GVBUJQXJZASXWHTWCLN2WNT57E",
        "ReplicationType": "full-load",
    }
}

MOCK_DESCRIBE_REPLICATIONS_RESP = {
    "Marker": "string",
    "Replications": [
        {
            "ReplicationConfigIdentifier": "test-config",
            "ReplicationConfigArn": "arn:aws:dms:us-east-1:123456789012:replication-config/test-config",
            "SourceEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:RZZK4EZW5UANC7Y3P4E776WHBE",
            "TargetEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:GVBUJQXJZASXWHTWCLN2WNT57E",
            "ReplicationType": "full-load",
            "Status": "CREATED",
            "ProvisionData": {
                "ProvisionState": "string",
                "ProvisionedCapacityUnits": 123,
                "DateProvisioned": datetime(2015, 1, 1),
                "IsNewProvisioningAvailable": False,
                "DateNewProvisioningDataAvailable": datetime(2015, 1, 1),
                "ReasonForNewProvisioningData": "string",
            },
            "StopReason": "string",
            "FailureMessages": [
                "string",
            ],
            "ReplicationStats": {
                "FullLoadProgressPercent": 123,
                "ElapsedTimeMillis": 123,
                "TablesLoaded": 123,
                "TablesLoading": 123,
                "TablesQueued": 123,
                "TablesErrored": 123,
                "FreshStartDate": datetime(2015, 1, 1),
                "StartDate": datetime(2015, 1, 1),
                "StopDate": datetime(2015, 1, 1),
                "FullLoadStartDate": datetime(2015, 1, 1),
                "FullLoadFinishDate": datetime(2015, 1, 1),
            },
            "StartReplicationType": "string",
            "CdcStartTime": datetime(2015, 1, 1),
            "CdcStartPosition": "string",
            "CdcStopPosition": "string",
            "RecoveryCheckpoint": "string",
            "ReplicationCreateTime": datetime(2015, 1, 1),
            "ReplicationUpdateTime": datetime(2015, 1, 1),
            "ReplicationLastStopTime": datetime(2015, 1, 1),
            "ReplicationDeprovisionTime": datetime(2015, 1, 1),
        },
        {
            "ReplicationConfigIdentifier": "test-config-2",
            "ReplicationConfigArn": "arn:aws:dms:us-east-1:123456789012:replication-config/test-config-2",
            "SourceEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:RZZK4EZW5UANC7Y3P4E776WHBE",
            "TargetEndpointArn": "arn:aws:dms:us-east-1:123456789012:endpoint:GVBUJQXJZASXWHTWCLN2WNT57E",
            "ReplicationType": "cdc-only",
            "Status": "CREATED",
            "ProvisionData": {
                "ProvisionState": "string",
                "ProvisionedCapacityUnits": 123,
                "DateProvisioned": datetime(2015, 1, 1),
                "IsNewProvisioningAvailable": False,
                "DateNewProvisioningDataAvailable": datetime(2015, 1, 1),
                "ReasonForNewProvisioningData": "string",
            },
            "StopReason": "string",
            "FailureMessages": [
                "string",
            ],
            "ReplicationStats": {
                "FullLoadProgressPercent": 123,
                "ElapsedTimeMillis": 123,
                "TablesLoaded": 123,
                "TablesLoading": 123,
                "TablesQueued": 123,
                "TablesErrored": 123,
                "FreshStartDate": datetime(2015, 1, 1),
                "StartDate": datetime(2015, 1, 1),
                "StopDate": datetime(2015, 1, 1),
                "FullLoadStartDate": datetime(2015, 1, 1),
                "FullLoadFinishDate": datetime(2015, 1, 1),
            },
            "StartReplicationType": "string",
            "CdcStartTime": datetime(2015, 1, 1),
            "CdcStartPosition": "string",
            "CdcStopPosition": "string",
            "RecoveryCheckpoint": "string",
            "ReplicationCreateTime": datetime(2015, 1, 1),
            "ReplicationUpdateTime": datetime(2015, 1, 1),
            "ReplicationLastStopTime": datetime(2015, 1, 1),
            "ReplicationDeprovisionTime": datetime(2015, 1, 1),
        },
    ],
}


class TestDmsHook:
    def setup_method(self):
        self.dms = DmsHook()

    def test_init(self):
        assert self.dms.aws_conn_id == "aws_default"

    @mock.patch.object(DmsHook, "get_conn")
    def test_describe_replication_tasks_with_no_tasks_found(self, mock_conn):
        mock_conn.return_value.describe_replication_tasks.return_value = {}

        marker, tasks = self.dms.describe_replication_tasks()

        mock_conn.return_value.describe_replication_tasks.assert_called_once()
        assert marker is None
        assert len(tasks) == 0

    @mock.patch.object(DmsHook, "get_conn")
    def test_describe_replication_tasks(self, mock_conn):
        mock_conn.return_value.describe_replication_tasks.return_value = MOCK_DESCRIBE_RESPONSE
        describe_tasks_kwargs = {
            "Filters": [{"Name": "replication-task-id", "Values": [MOCK_DATA["replication_task_id"]]}]
        }

        marker, tasks = self.dms.describe_replication_tasks(**describe_tasks_kwargs)

        mock_conn.return_value.describe_replication_tasks.assert_called_with(**describe_tasks_kwargs)
        assert marker is None
        assert len(tasks) == 1
        assert tasks[0]["ReplicationTaskArn"] == MOCK_TASK_ARN

    @mock.patch.object(DmsHook, "get_conn")
    def test_describe_teplication_tasks_with_marker(self, mock_conn):
        mock_conn.return_value.describe_replication_tasks.return_value = MOCK_DESCRIBE_RESPONSE_WITH_MARKER
        describe_tasks_kwargs = {
            "Filters": [{"Name": "replication-task-id", "Values": [MOCK_DATA["replication_task_id"]]}]
        }

        marker, tasks = self.dms.describe_replication_tasks(**describe_tasks_kwargs)

        mock_conn.return_value.describe_replication_tasks.assert_called_with(**describe_tasks_kwargs)
        assert marker == MOCK_DESCRIBE_RESPONSE_WITH_MARKER["Marker"]
        assert len(tasks) == 1
        assert tasks[0]["ReplicationTaskArn"] == MOCK_TASK_ARN

    @mock.patch.object(DmsHook, "get_conn")
    def test_find_replication_tasks_by_arn(self, mock_conn):
        mock_conn.return_value.describe_replication_tasks.return_value = MOCK_DESCRIBE_RESPONSE

        tasks = self.dms.find_replication_tasks_by_arn(MOCK_TASK_ARN)

        expected_call_params = {
            "Filters": [{"Name": "replication-task-arn", "Values": [MOCK_TASK_ARN]}],
            "WithoutSettings": False,
        }

        mock_conn.return_value.describe_replication_tasks.assert_called_with(**expected_call_params)
        assert len(tasks) == 1
        assert tasks[0]["ReplicationTaskArn"] == MOCK_TASK_ARN

    @mock.patch.object(DmsHook, "get_conn")
    def test_get_task_status(self, mock_conn):
        mock_conn.return_value.describe_replication_tasks.return_value = MOCK_DESCRIBE_RESPONSE

        status = self.dms.get_task_status(MOCK_TASK_ARN)

        expected_call_params = {
            "Filters": [{"Name": "replication-task-arn", "Values": [MOCK_TASK_ARN]}],
            "WithoutSettings": True,
        }

        mock_conn.return_value.describe_replication_tasks.assert_called_with(**expected_call_params)
        assert status == MOCK_TASK_RESPONSE_DATA["Status"]

    @mock.patch.object(DmsHook, "get_conn")
    def test_get_task_status_with_no_task_found(self, mock_conn):
        mock_conn.return_value.describe_replication_tasks.return_value = {}
        status = self.dms.get_task_status(MOCK_TASK_ARN)

        mock_conn.return_value.describe_replication_tasks.assert_called_once()
        assert status is None

    @mock.patch.object(DmsHook, "get_conn")
    def test_create_replication_task(self, mock_conn):
        mock_conn.return_value.create_replication_task.return_value = MOCK_CREATE_RESPONSE
        result = self.dms.create_replication_task(**MOCK_DATA)
        expected_call_params = {
            "ReplicationTaskIdentifier": MOCK_DATA["replication_task_id"],
            "SourceEndpointArn": MOCK_DATA["source_endpoint_arn"],
            "TargetEndpointArn": MOCK_DATA["target_endpoint_arn"],
            "ReplicationInstanceArn": MOCK_DATA["replication_instance_arn"],
            "MigrationType": MOCK_DATA["migration_type"],
            "TableMappings": json.dumps(MOCK_DATA["table_mappings"]),
        }
        mock_conn.return_value.create_replication_task.assert_called_with(**expected_call_params)
        assert result == MOCK_CREATE_RESPONSE["ReplicationTask"]["ReplicationTaskArn"]

    @mock.patch.object(DmsHook, "get_conn")
    def test_start_replication_task(self, mock_conn):
        mock_conn.return_value.start_replication_task.return_value = MOCK_START_RESPONSE
        start_type = "start-replication"

        self.dms.start_replication_task(
            replication_task_arn=MOCK_TASK_ARN,
            start_replication_task_type=start_type,
        )

        expected_call_params = {
            "ReplicationTaskArn": MOCK_TASK_ARN,
            "StartReplicationTaskType": start_type,
        }
        mock_conn.return_value.start_replication_task.assert_called_with(**expected_call_params)

    @mock.patch.object(DmsHook, "get_conn")
    def test_stop_replication_task(self, mock_conn):
        mock_conn.return_value.stop_replication_task.return_value = MOCK_STOP_RESPONSE

        self.dms.stop_replication_task(replication_task_arn=MOCK_TASK_ARN)

        expected_call_params = {"ReplicationTaskArn": MOCK_TASK_ARN}
        mock_conn.return_value.stop_replication_task.assert_called_with(**expected_call_params)

    @mock.patch.object(DmsHook, "get_conn")
    def test_delete_replication_task(self, mock_conn):
        mock_conn.return_value.delete_replication_task.return_value = MOCK_DELETE_RESPONSE

        self.dms.delete_replication_task(replication_task_arn=MOCK_TASK_ARN)

        expected_call_params = {"ReplicationTaskArn": MOCK_TASK_ARN}
        mock_conn.return_value.delete_replication_task.assert_called_with(**expected_call_params)

    @mock.patch.object(DmsHook, "get_conn")
    def test_wait_for_task_status_with_unknown_target_status(self, mock_conn):
        with pytest.raises(TypeError, match="Status must be an instance of DmsTaskWaiterStatus"):
            self.dms.wait_for_task_status(MOCK_TASK_ARN, "unknown_status")

    @mock.patch.object(DmsHook, "get_conn")
    def test_wait_for_task_status(self, mock_conn):
        self.dms.wait_for_task_status(replication_task_arn=MOCK_TASK_ARN, status=DmsTaskWaiterStatus.DELETED)

        expected_waiter_call_params = {
            "Filters": [{"Name": "replication-task-arn", "Values": [MOCK_TASK_ARN]}],
            "WithoutSettings": True,
        }
        mock_conn.return_value.get_waiter.assert_called_with("replication_task_deleted")
        mock_conn.return_value.get_waiter.return_value.wait.assert_called_with(**expected_waiter_call_params)

    @mock.patch.object(DmsHook, "conn")
    def test_describe_config_no_filter(self, mock_conn):
        mock_conn.describe_replication_configs.return_value = MOCK_CONFIG_RESPONSE
        resp = self.dms.describe_replication_configs()

        assert len(resp) == 2

    @mock.patch.object(DmsHook, "conn")
    def test_describe_config_filter(self, mock_conn):
        filter = [{"Name": "replication-type", "Values": ["cdc"]}]
        self.dms.describe_replication_configs(filters=filter)
        mock_conn.describe_replication_configs.assert_called_with(Filters=filter)

    @mock.patch.object(DmsHook, "conn")
    def test_create_repl_config_kwargs(self, mock_conn):
        self.dms.create_replication_config(
            replication_config_id=MOCK_REPLICATION_CONFIG["ReplicationConfigIdentifier"],
            source_endpoint_arn=MOCK_REPLICATION_CONFIG["SourceEndpointArn"],
            target_endpoint_arn=MOCK_REPLICATION_CONFIG["TargetEndpointArn"],
            compute_config=MOCK_REPLICATION_CONFIG["ComputeConfig"],
            replication_type=MOCK_REPLICATION_CONFIG["ReplicationType"],
            table_mappings=MOCK_REPLICATION_CONFIG["TableMappings"],
            additional_config_kwargs={
                "ReplicationSettings": MOCK_REPLICATION_CONFIG["ReplicationSettings"],
                "SupplementalSettings": MOCK_REPLICATION_CONFIG["SupplementalSettings"],
            },
        )

        mock_conn.create_replication_config.assert_called_with(
            ReplicationConfigIdentifier=MOCK_REPLICATION_CONFIG["ReplicationConfigIdentifier"],
            SourceEndpointArn=MOCK_REPLICATION_CONFIG["SourceEndpointArn"],
            TargetEndpointArn=MOCK_REPLICATION_CONFIG["TargetEndpointArn"],
            ComputeConfig=MOCK_REPLICATION_CONFIG["ComputeConfig"],
            ReplicationType=MOCK_REPLICATION_CONFIG["ReplicationType"],
            TableMappings=MOCK_REPLICATION_CONFIG["TableMappings"],
            ReplicationSettings=MOCK_REPLICATION_CONFIG["ReplicationSettings"],
            SupplementalSettings=MOCK_REPLICATION_CONFIG["SupplementalSettings"],
        )

        self.dms.create_replication_config(
            replication_config_id=MOCK_REPLICATION_CONFIG["ReplicationConfigIdentifier"],
            source_endpoint_arn=MOCK_REPLICATION_CONFIG["SourceEndpointArn"],
            target_endpoint_arn=MOCK_REPLICATION_CONFIG["TargetEndpointArn"],
            compute_config=MOCK_REPLICATION_CONFIG["ComputeConfig"],
            replication_type=MOCK_REPLICATION_CONFIG["ReplicationType"],
            table_mappings=MOCK_REPLICATION_CONFIG["TableMappings"],
        )
        mock_conn.create_replication_config.assert_called_with(
            ReplicationConfigIdentifier=MOCK_REPLICATION_CONFIG["ReplicationConfigIdentifier"],
            SourceEndpointArn=MOCK_REPLICATION_CONFIG["SourceEndpointArn"],
            TargetEndpointArn=MOCK_REPLICATION_CONFIG["TargetEndpointArn"],
            ComputeConfig=MOCK_REPLICATION_CONFIG["ComputeConfig"],
            ReplicationType=MOCK_REPLICATION_CONFIG["ReplicationType"],
            TableMappings=MOCK_REPLICATION_CONFIG["TableMappings"],
        )

    @mock.patch.object(DmsHook, "conn")
    def test_create_repl_config(self, mock_conn):
        mock_conn.create_replication_config.return_value = MOCK_REPLICATION_CONFIG_RESP

        resp = self.dms.create_replication_config(
            replication_config_id=MOCK_REPLICATION_CONFIG["ReplicationConfigIdentifier"],
            source_endpoint_arn=MOCK_REPLICATION_CONFIG["SourceEndpointArn"],
            target_endpoint_arn=MOCK_REPLICATION_CONFIG["TargetEndpointArn"],
            compute_config=MOCK_REPLICATION_CONFIG["ComputeConfig"],
            replication_type=MOCK_REPLICATION_CONFIG["ReplicationType"],
            table_mappings=MOCK_REPLICATION_CONFIG["TableMappings"],
        )

        assert resp == MOCK_REPLICATION_CONFIG_RESP["ReplicationConfig"]["ReplicationConfigArn"]

    @mock.patch.object(DmsHook, "conn")
    def test_describe_replications(self, mock_conn):
        mock_conn.describe_replication_tasks.return_value = MOCK_DESCRIBE_REPLICATIONS_RESP
        resp = self.dms.describe_replication_tasks()
        assert len(resp) == 2

    @mock.patch.object(DmsHook, "conn")
    def test_describe_replications_filter(self, mock_conn):
        filter = [
            {
                "Name": "replication-task-id",
                "Values": MOCK_DESCRIBE_REPLICATIONS_RESP["Replications"][0]["ReplicationConfigArn"],
            }
        ]
        self.dms.describe_replication_tasks(filters=filter)
        mock_conn.describe_replication_tasks.assert_called_with(filters=filter)

    @mock.patch.object(DmsHook, "conn")
    def test_start_replication_args(self, mock_conn):
        self.dms.start_replication(
            replication_config_arn=MOCK_TASK_ARN,
            start_replication_type="cdc",
        )
        mock_conn.start_replication.assert_called_with(
            ReplicationConfigArn=MOCK_TASK_ARN,
            StartReplicationType="cdc",
        )

    @mock.patch.object(DmsHook, "conn")
    def test_start_replication_kwargs(self, mock_conn):
        self.dms.start_replication(
            replication_config_arn=MOCK_TASK_ARN,
            start_replication_type="cdc",
            cdc_start_time="2022-01-01T00:00:00Z",
            cdc_start_pos=None,
            cdc_stop_pos=None,
        )
        mock_conn.start_replication.assert_called_with(
            ReplicationConfigArn=MOCK_TASK_ARN,
            StartReplicationType="cdc",
            CdcStartTime=parser.parse("2022-01-01T00:00:00Z"),
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.sns import SnsHook

DEDUPE_ID = "test-dedupe-id"
GROUP_ID = "test-group-id"
MESSAGE = "Hello world"
SUBJECT = "test-subject"
INVALID_ATTRIBUTES_MSG = r"Values in MessageAttributes must be one of bytes, str, int, float, or iterable"

TOPIC_NAME = "test-topic"
TOPIC_ARN = f"arn:aws:sns:us-east-1:123456789012:{TOPIC_NAME}"

INVALID_ATTRIBUTES = {"test-non-iterable": object()}
VALID_ATTRIBUTES = {
    "test-string": "string-value",
    "test-number": 123456,
    "test-array": ["first", "second", "third"],
    "test-binary": b"binary-value",
}

MESSAGE_ID_KEY = "MessageId"
TOPIC_ARN_KEY = "TopicArn"


class TestSnsHook:
    @pytest.fixture(autouse=True)
    def setup_moto(self):
        with mock_aws():
            yield

    @pytest.fixture
    def hook(self):
        return SnsHook(aws_conn_id="aws_default")

    @pytest.fixture
    def target(self, hook):
        return hook.get_conn().create_topic(Name=TOPIC_NAME).get(TOPIC_ARN_KEY)

    def test_get_conn_returns_a_boto3_connection(self, hook):
        assert hook.get_conn() is not None

    def test_publish_to_target_with_subject(self, hook, target):
        response = hook.publish_to_target(target, MESSAGE, SUBJECT)

        assert MESSAGE_ID_KEY in response

    def test_publish_to_target_with_attributes(self, hook, target):
        response = hook.publish_to_target(target, MESSAGE, message_attributes=VALID_ATTRIBUTES)

        assert MESSAGE_ID_KEY in response

    def test_publish_to_target_plain(self, hook, target):
        response = hook.publish_to_target(target, MESSAGE)

        assert MESSAGE_ID_KEY in response

    def test_publish_to_target_error(self, hook, target):
        with pytest.raises(TypeError, match=INVALID_ATTRIBUTES_MSG):
            hook.publish_to_target(target, MESSAGE, message_attributes=INVALID_ATTRIBUTES)

    def test_publish_to_target_with_deduplication(self, hook):
        fifo_target = (
            hook.get_conn()
            .create_topic(
                Name=f"{TOPIC_NAME}.fifo",
                Attributes={
                    "FifoTopic": "true",
                    "ContentBasedDeduplication": "false",
                },
            )
            .get("TopicArn")
        )

        response = hook.publish_to_target(
            fifo_target, MESSAGE, message_deduplication_id=DEDUPE_ID, message_group_id=GROUP_ID
        )
        assert MESSAGE_ID_KEY in response


@pytest.mark.asyncio
class TestAsyncSnsHook:
    """The mock_aws decorator uses `moto` which does not currently support async SNS so we mock it manually."""

    @pytest.fixture
    def hook(self):
        return SnsHook(aws_conn_id="aws_default")

    @pytest.fixture
    def mock_async_client(self):
        mock_client = mock.AsyncMock()
        mock_client.publish.return_value = {MESSAGE_ID_KEY: "test-message-id"}
        return mock_client

    @pytest.fixture
    def mock_get_async_conn(self, mock_async_client):
        with mock.patch.object(SnsHook, "get_async_conn") as mocked_conn:
            mocked_conn.return_value = mock_async_client
            mocked_conn.return_value.__aenter__.return_value = mock_async_client
            yield mocked_conn

    async def test_get_async_conn(self, hook, mock_get_async_conn, mock_async_client):
        # Test context manager access
        async with await hook.get_async_conn() as async_conn:
            assert async_conn is mock_async_client

        # Test direct access
        async_conn = await hook.get_async_conn()
        assert async_conn is mock_async_client

    async def test_apublish_to_target_with_subject(self, hook, mock_get_async_conn, mock_async_client):
        response = await hook.apublish_to_target(TOPIC_ARN, MESSAGE, SUBJECT)

        assert MESSAGE_ID_KEY in response

    async def test_apublish_to_target_with_attributes(self, hook, mock_get_async_conn, mock_async_client):
        response = await hook.apublish_to_target(TOPIC_ARN, MESSAGE, message_attributes=VALID_ATTRIBUTES)

        assert MESSAGE_ID_KEY in response

    async def test_publish_to_target_plain(self, hook, mock_get_async_conn, mock_async_client):
        response = await hook.apublish_to_target(TOPIC_ARN, MESSAGE)

        assert MESSAGE_ID_KEY in response

    async def test_publish_to_target_error(self, hook, mock_get_async_conn, mock_async_client):
        with pytest.raises(TypeError, match=INVALID_ATTRIBUTES_MSG):
            await hook.apublish_to_target(TOPIC_ARN, MESSAGE, message_attributes=INVALID_ATTRIBUTES)

    async def test_apublish_to_target_with_deduplication(self, hook, mock_get_async_conn, mock_async_client):
        response = await hook.apublish_to_target(
            TOPIC_ARN, MESSAGE, message_deduplication_id=DEDUPE_ID, message_group_id=GROUP_ID
        )

        assert MESSAGE_ID_KEY in response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.amazon.aws.hooks.mwaa import MwaaHook
from airflow.providers.amazon.aws.triggers.mwaa import MwaaDagRunCompletedTrigger, MwaaTaskCompletedTrigger
from airflow.triggers.base import TriggerEvent
from airflow.utils.state import DagRunState

from unit.amazon.aws.utils.test_waiter import assert_expected_waiter_type

BASE_TRIGGER_CLASSPATH = "airflow.providers.amazon.aws.triggers.mwaa."
TRIGGER_DAG_RUN_KWARGS = {
    "external_env_name": "test_env",
    "external_dag_id": "test_dag",
    "external_dag_run_id": "test_run_id",
}

TRIGGER_TASK_KWARGS = {
    "external_env_name": "test_env",
    "external_dag_id": "test_dag",
    "external_dag_run_id": "test_run_id",
    "external_task_id": "test_task_id",
}


class TestMwaaDagRunCompletedTrigger:
    def test_init_states(self):
        trigger = MwaaDagRunCompletedTrigger(**TRIGGER_DAG_RUN_KWARGS)
        assert trigger.success_states == {DagRunState.SUCCESS.value}
        assert trigger.failure_states == {DagRunState.FAILED.value}
        acceptors = trigger.waiter_config_overrides["acceptors"]
        expected_acceptors = [
            {
                "matcher": "path",
                "argument": "RestApiResponse.state",
                "expected": DagRunState.SUCCESS.value,
                "state": "success",
            },
            {
                "matcher": "path",
                "argument": "RestApiResponse.state",
                "expected": DagRunState.FAILED.value,
                "state": "failure",
            },
            {
                "matcher": "path",
                "argument": "RestApiResponse.state",
                "expected": DagRunState.RUNNING.value,
                "state": "retry",
            },
            {
                "matcher": "path",
                "argument": "RestApiResponse.state",
                "expected": DagRunState.QUEUED.value,
                "state": "retry",
            },
        ]
        assert len(acceptors) == len(DagRunState)
        assert {tuple(sorted(a.items())) for a in acceptors} == {
            tuple(sorted(a.items())) for a in expected_acceptors
        }

    def test_init_fail(self):
        with pytest.raises(ValueError, match=r".*success_states.*failure_states.*"):
            MwaaDagRunCompletedTrigger(
                **TRIGGER_DAG_RUN_KWARGS, success_states=("a", "b"), failure_states=("b", "c")
            )

    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = MwaaDagRunCompletedTrigger(**TRIGGER_DAG_RUN_KWARGS, aws_conn_id=OVERWRITTEN_CONN)
        assert op.hook().aws_conn_id == OVERWRITTEN_CONN

    def test_no_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = MwaaDagRunCompletedTrigger(**TRIGGER_DAG_RUN_KWARGS)
        assert op.hook().aws_conn_id == DEFAULT_CONN

    def test_serialization(self):
        success_states = ["a", "b"]
        failure_states = ["c", "d"]
        trigger = MwaaDagRunCompletedTrigger(
            **TRIGGER_DAG_RUN_KWARGS, success_states=success_states, failure_states=failure_states
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "MwaaDagRunCompletedTrigger"
        assert kwargs.get("external_env_name") == TRIGGER_DAG_RUN_KWARGS["external_env_name"]
        assert kwargs.get("external_dag_id") == TRIGGER_DAG_RUN_KWARGS["external_dag_id"]
        assert kwargs.get("external_dag_run_id") == TRIGGER_DAG_RUN_KWARGS["external_dag_run_id"]
        assert kwargs.get("success_states") == success_states
        assert kwargs.get("failure_states") == failure_states

    @pytest.mark.asyncio
    @mock.patch.object(MwaaHook, "get_waiter")
    @mock.patch.object(MwaaHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = MwaaDagRunCompletedTrigger(**TRIGGER_DAG_RUN_KWARGS)

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent(
            {"status": "success", "dag_run_id": TRIGGER_DAG_RUN_KWARGS["external_dag_run_id"]}
        )
        assert_expected_waiter_type(mock_get_waiter, "mwaa_dag_run_complete")
        mock_get_waiter().wait.assert_called_once()


class TestMwaaTaskCompletedTrigger:
    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = MwaaTaskCompletedTrigger(**TRIGGER_TASK_KWARGS, aws_conn_id=OVERWRITTEN_CONN)
        assert op.hook().aws_conn_id == OVERWRITTEN_CONN

    def test_no_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = MwaaTaskCompletedTrigger(**TRIGGER_TASK_KWARGS)
        assert op.hook().aws_conn_id == DEFAULT_CONN

    def test_init_fail(self):
        with pytest.raises(ValueError, match=r".*success_states.*failure_states.*"):
            MwaaTaskCompletedTrigger(
                **TRIGGER_TASK_KWARGS, success_states=("a", "b"), failure_states=("b", "c")
            )

    def test_serialization(self):
        success_states = ["a", "b"]
        failure_states = ["c", "d"]
        trigger = MwaaTaskCompletedTrigger(
            **TRIGGER_TASK_KWARGS, success_states=success_states, failure_states=failure_states
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "MwaaTaskCompletedTrigger"
        assert kwargs.get("external_env_name") == TRIGGER_TASK_KWARGS["external_env_name"]
        assert kwargs.get("external_dag_id") == TRIGGER_TASK_KWARGS["external_dag_id"]
        assert kwargs.get("external_dag_run_id") == TRIGGER_TASK_KWARGS["external_dag_run_id"]
        assert kwargs.get("external_task_id") == TRIGGER_TASK_KWARGS["external_task_id"]
        assert kwargs.get("success_states") == success_states
        assert kwargs.get("failure_states") == failure_states

    @pytest.mark.asyncio
    @mock.patch.object(MwaaHook, "get_waiter")
    @mock.patch.object(MwaaHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = MwaaTaskCompletedTrigger(**TRIGGER_TASK_KWARGS)

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent(
            {"status": "success", "task_id": TRIGGER_TASK_KWARGS["external_task_id"]}
        )
        assert_expected_waiter_type(mock_get_waiter, "mwaa_task_complete")
        mock_get_waiter().wait.assert_called_once()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.athena import AthenaHook
from airflow.providers.amazon.aws.sensors.athena import AthenaSensor
from airflow.providers.common.compat.sdk import AirflowException


@pytest.fixture
def mock_poll_query_status():
    with mock.patch.object(AthenaHook, "poll_query_status") as m:
        yield m


@mock_aws
class TestAthenaSensor:
    def setup_method(self, _):
        self.default_op_kwargs = dict(
            task_id="test_athena_sensor",
            query_execution_id="abc",
            sleep_time=5,
            max_retries=1,
        )
        self.sensor = AthenaSensor(**self.default_op_kwargs, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = AthenaSensor(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None
        assert op.hook.log_query is True

        op = AthenaSensor(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @pytest.mark.parametrize("state", ["SUCCEEDED"])
    def test_poke_success_states(self, state, mock_poll_query_status):
        mock_poll_query_status.side_effect = [state]
        assert self.sensor.poke({}) is True

    @pytest.mark.parametrize("state", ["RUNNING", "QUEUED"])
    def test_poke_intermediate_states(self, state, mock_poll_query_status):
        mock_poll_query_status.side_effect = [state]
        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", ["FAILED", "CANCELLED"])
    def test_poke_failure_states(self, state, mock_poll_query_status):
        mock_poll_query_status.side_effect = [state]
        sensor = AthenaSensor(**self.default_op_kwargs, aws_conn_id=None)
        message = "Athena sensor failed"
        with pytest.raises(AirflowException, match=message):
            sensor.poke({})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from enum import Enum
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.glacier import GlacierHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException

if TYPE_CHECKING:
    from airflow.sdk import Context


class JobStatus(Enum):
    """Glacier jobs description."""

    IN_PROGRESS = "InProgress"
    SUCCEEDED = "Succeeded"


class GlacierJobOperationSensor(AwsBaseSensor[GlacierHook]):
    """
    Glacier sensor for checking job state. This operator runs only in reschedule mode.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:GlacierJobOperationSensor`

    :param aws_conn_id: The reference to the AWS connection details
    :param vault_name: name of Glacier vault on which job is executed
    :param job_id: the job ID was returned by retrieve_inventory()
    :param poke_interval: Time in seconds that the job should wait in
        between each tries
    :param mode: How the sensor operates.
        Options are: ``{ poke | reschedule }``, default is ``poke``.
        When set to ``poke`` the sensor is taking up a worker slot for its
        whole execution time and sleeps between pokes. Use this mode if the
        expected runtime of the sensor is short or if a short poke interval
        is required. Note that the sensor will hold onto a worker slot and
        a pool slot for the duration of the sensor's runtime in this mode.
        When set to ``reschedule`` the sensor task frees the worker slot when
        the criteria is not yet met and it's rescheduled at a later time. Use
        this mode if the time before the criteria is met is expected to be
        quite long. The poke interval should be more than one minute to
        prevent too much load on the scheduler.
    """

    aws_hook_class = GlacierHook
    template_fields: Sequence[str] = aws_template_fields("vault_name", "job_id")

    def __init__(
        self,
        *,
        vault_name: str,
        job_id: str,
        poke_interval: int = 60 * 20,
        mode: str = "reschedule",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.vault_name = vault_name
        self.job_id = job_id
        self.poke_interval = poke_interval
        self.mode = mode

    def poke(self, context: Context) -> bool:
        response = self.hook.describe_job(vault_name=self.vault_name, job_id=self.job_id)

        if response["StatusCode"] == JobStatus.SUCCEEDED.value:
            self.log.info("Job status: %s, code status: %s", response["Action"], response["StatusCode"])
            self.log.info("Job finished successfully")
            return True
        if response["StatusCode"] == JobStatus.IN_PROGRESS.value:
            self.log.info("Processing...")
            self.log.warning("Code status: %s", response["StatusCode"])
            return False
        raise AirflowException(
            f"Sensor failed. Job status: {response['Action']}, code status: {response['StatusCode']}"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import logging
import os
import subprocess
from tempfile import mkdtemp

import boto3

"""
Example Lambda function to execute an Airflow command or workload. Use or modify this code as needed.
"""

log = logging.getLogger()
log.setLevel(logging.INFO)


# Get the S3 URI from the environment variable. Set either on the Lambda function or in the
# docker image used for the lambda invocations.
S3_URI = os.environ.get("S3_URI", None)
# Input and output keys
TASK_KEY_KEY = "task_key"
COMMAND_KEY = "command"
EXECUTOR_CONFIG_KEY = "executor_config"
RETURN_CODE_KEY = "return_code"


def lambda_handler(event, context):
    log.info("Received event: %s", event)
    log.info("Received context: %s", context)

    command = event.get(COMMAND_KEY)
    task_key = event.get(TASK_KEY_KEY)
    executor_config = event.get(EXECUTOR_CONFIG_KEY, {})  # noqa: F841

    # Any pre-processing or validation of the command or use of the executor_config can be done here or above.

    # Sync dags from s3 to the local dags directory
    if S3_URI:
        fetch_dags_from_s3(S3_URI)
    # This function must be called, it executes the Airflow command and reports to SQS.
    run_and_report(command, task_key)

    # Any post-processing or cleanup can be done here.


def run_and_report(command, task_key):
    """Execute the provided Airflow command or workload and report the result via SQS."""
    try:
        log.info("Starting execution for task: %s", task_key)
        result = subprocess.run(
            command,
            check=False,
            shell=isinstance(command, str),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        return_code = result.returncode
        log.info("Execution completed for task %s with return code %s", task_key, return_code)
        log.info("Output:")
        log.info("%s", result.stdout.decode())
    except Exception:
        log.exception("Error executing task %s: ", task_key)
        return_code = 1  # Non-zero indicates failure to run the task

    queue_url = get_queue_url()
    message = json.dumps({TASK_KEY_KEY: task_key, RETURN_CODE_KEY: return_code})
    try:
        sqs_client = get_sqs_client()
        sqs_client.send_message(QueueUrl=queue_url, MessageBody=message)
        log.info("Sent result to SQS %s", message)
    except Exception:
        log.exception("Failed to send message to SQS for task %s", task_key)


def get_sqs_client():
    """Create an SQS client. Credentials and region are automatically picked up from the environment."""
    return boto3.client("sqs")


def get_queue_url():
    """
    Get the SQS queue URL from the environment variable.

    Set either on the Lambda function or in the image used for the lambda invocations.
    """
    queue_url = os.environ.get("AIRFLOW__AWS_LAMBDA_EXECUTOR__QUEUE_URL", os.environ.get("QUEUE_URL", None))
    if not queue_url:
        raise RuntimeError(
            "No Queue URL detected (either AIRFLOW__AWS_LAMBDA_EXECUTOR__QUEUE_URL or "
            "QUEUE_URL); Will be unable to send task results. Exiting!"
        )
    return queue_url


def fetch_dags_from_s3(s3_uri):
    """Fetch DAGs from S3 and sync them to the local dags directory."""
    log.info("Fetching DAGs from S3 URI: %s", s3_uri)
    # Use a named temporary directory for the local dags folder, only tmp is writeable in Lambda
    local_dags_dir = mkdtemp(prefix="airflow_dags_")
    log.info("Setting AIRFLOW__CORE__DAGS_FOLDER to: %s", local_dags_dir)
    os.environ["AIRFLOW__CORE__DAGS_FOLDER"] = local_dags_dir

    # S3 URI format s3://bucket-name/path/to/dags/
    bucket_name = s3_uri.split("/")[2]
    prefix = "/".join(s3_uri.split("/")[3:])

    s3_resource = boto3.resource("s3")
    bucket = s3_resource.Bucket(bucket_name)

    for obj in bucket.objects.filter(Prefix=prefix):
        if obj.key.endswith("/"):
            # Skip directories
            continue
        key = obj.key
        local_path = os.path.join(local_dags_dir, os.path.basename(key))
        log.info("Downloading %s to %s", key, local_path)
        bucket.download_file(key, local_path)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.notifications.sqs import SqsNotifier, send_sqs_notification
from airflow.providers.amazon.version_compat import NOTSET

PARAM_DEFAULT_VALUE = pytest.param(NOTSET, id="default-value")

SEND_MSG_KWARGS = {
    "queue_url": "https://sqs.eu-west-1.amazonaws.com/123456789098/MyQueue",
    "message_body": "foo-bar",
    "delay_seconds": 42,
    "message_attributes": {},
    "message_group_id": "foo-bar",
}


class TestSqsNotifier:
    def test_class_and_notifier_are_same(self):
        assert send_sqs_notification is SqsNotifier

    @pytest.mark.parametrize("aws_conn_id", ["aws_test_conn_id", None, PARAM_DEFAULT_VALUE])
    @pytest.mark.parametrize("region_name", ["eu-west-2", None, PARAM_DEFAULT_VALUE])
    def test_parameters_propagate_to_hook(self, aws_conn_id, region_name):
        """Test notifier attributes propagate to SqsHook."""

        notifier_kwargs = {}
        if aws_conn_id is not NOTSET:
            notifier_kwargs["aws_conn_id"] = aws_conn_id
        if region_name is not NOTSET:
            notifier_kwargs["region_name"] = region_name

        notifier = SqsNotifier(**notifier_kwargs, **SEND_MSG_KWARGS)
        with mock.patch("airflow.providers.amazon.aws.notifications.sqs.SqsHook") as mock_hook:
            hook = notifier.hook
            assert hook is notifier.hook, "Hook property not cached"
            mock_hook.assert_called_once_with(
                aws_conn_id=(aws_conn_id if aws_conn_id is not NOTSET else "aws_default"),
                region_name=(region_name if region_name is not NOTSET else None),
            )

            # Basic check for notifier
            notifier.notify({})
            mock_hook.return_value.send_message.assert_called_once_with(**SEND_MSG_KWARGS)

    def test_sqs_notifier_templated(self, create_dag_without_db):
        notifier = SqsNotifier(
            aws_conn_id="{{ dag.dag_id }}",
            queue_url="https://sqs.{{ var_region }}.amazonaws.com/{{ var_account }}/{{ var_queue }}",
            message_body="The {{ var_username|capitalize }} Show",
            message_attributes={"bar": "{{ dag.dag_id }}"},
            message_group_id="{{ var_group_id }}",
            region_name="{{ var_region }}",
        )
        with mock.patch("airflow.providers.amazon.aws.notifications.sqs.SqsHook") as m:
            notifier(
                {
                    "dag": create_dag_without_db("test_sqs_notifier_templated"),
                    "var_username": "truman",
                    "var_region": "ca-central-1",
                    "var_account": "123321123321",
                    "var_queue": "AwesomeQueue",
                    "var_group_id": "spam",
                }
            )
            # Hook initialisation
            m.assert_called_once_with(aws_conn_id="test_sqs_notifier_templated", region_name="ca-central-1")
            # Send message
            m.return_value.send_message.assert_called_once_with(
                queue_url="https://sqs.ca-central-1.amazonaws.com/123321123321/AwesomeQueue",
                message_body="The Truman Show",
                message_group_id="spam",
                message_attributes={"bar": "test_sqs_notifier_templated"},
                delay_seconds=0,
            )

    @pytest.mark.asyncio
    async def test_async_notify(self):
        notifier = SqsNotifier(**SEND_MSG_KWARGS)
        with mock.patch("airflow.providers.amazon.aws.notifications.sqs.SqsHook") as mock_hook:
            mock_hook.return_value.asend_message = mock.AsyncMock()

            await notifier.async_notify({})

            mock_hook.return_value.asend_message.assert_called_once_with(**SEND_MSG_KWARGS)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from datetime import datetime

from airflow.providers.amazon.aws.operators.cloud_formation import (
    CloudFormationCreateStackOperator,
    CloudFormationDeleteStackOperator,
)
from airflow.providers.amazon.aws.sensors.cloud_formation import (
    CloudFormationCreateStackSensor,
    CloudFormationDeleteStackSensor,
)
from airflow.providers.common.compat.sdk import DAG, chain

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_cloudformation"

# The CloudFormation template must have at least one resource to
# be usable, this example uses SQS as a free and serverless option.
CLOUDFORMATION_TEMPLATE = {
    "Description": "Stack from Airflow CloudFormation example DAG",
    "Resources": {
        "ExampleQueue": {
            "Type": "AWS::SQS::Queue",
        }
    },
}

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    cloudformation_stack_name = f"{env_id}-stack"
    cloudformation_create_parameters = {
        "StackName": cloudformation_stack_name,
        "TemplateBody": json.dumps(CLOUDFORMATION_TEMPLATE),
        "TimeoutInMinutes": 2,
        "OnFailure": "DELETE",  # Don't leave stacks behind if creation fails.
    }

    # [START howto_operator_cloudformation_create_stack]
    create_stack = CloudFormationCreateStackOperator(
        task_id="create_stack",
        stack_name=cloudformation_stack_name,
        cloudformation_parameters=cloudformation_create_parameters,
    )
    # [END howto_operator_cloudformation_create_stack]

    # [START howto_sensor_cloudformation_create_stack]
    wait_for_stack_create = CloudFormationCreateStackSensor(
        task_id="wait_for_stack_create",
        stack_name=cloudformation_stack_name,
    )
    # [END howto_sensor_cloudformation_create_stack]
    wait_for_stack_create.poke_interval = 10

    # [START howto_operator_cloudformation_delete_stack]
    delete_stack = CloudFormationDeleteStackOperator(
        task_id="delete_stack",
        stack_name=cloudformation_stack_name,
    )
    # [END howto_operator_cloudformation_delete_stack]
    delete_stack.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_sensor_cloudformation_delete_stack]
    wait_for_stack_delete = CloudFormationDeleteStackSensor(
        task_id="wait_for_stack_delete",
        stack_name=cloudformation_stack_name,
    )
    # [END howto_sensor_cloudformation_delete_stack]
    wait_for_stack_delete.poke_interval = 10

    chain(
        # TEST SETUP
        test_context,
        # TEST BODY
        create_stack,
        wait_for_stack_create,
        delete_stack,
        wait_for_stack_delete,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime
from unittest import mock

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.step_function import StepFunctionHook
from airflow.providers.common.compat.sdk import AirflowFailException


@mock_aws
class TestStepFunctionHook:
    def test_get_conn_returns_a_boto3_connection(self):
        hook = StepFunctionHook(aws_conn_id="aws_default")
        assert hook.get_conn().meta.service_model.service_name == "stepfunctions"

    def test_start_execution(self):
        hook = StepFunctionHook(aws_conn_id="aws_default", region_name="us-east-1")
        state_machine = hook.get_conn().create_state_machine(
            name="pseudo-state-machine", definition="{}", roleArn="arn:aws:iam::000000000000:role/Role"
        )

        state_machine_arn = state_machine.get("stateMachineArn")

        execution_arn = hook.start_execution(
            state_machine_arn=state_machine_arn, name=None, state_machine_input={}
        )

        assert execution_arn is not None

    @mock.patch.object(StepFunctionHook, "conn")
    def test_redrive_execution(self, mock_conn):
        mock_conn.redrive_execution.return_value = {"redriveDate": datetime(2024, 1, 1)}
        StepFunctionHook().start_execution(
            state_machine_arn="arn:aws:states:us-east-1:123456789012:stateMachine:test-state-machine",
            name="random-123",
            is_redrive_execution=True,
        )

        mock_conn.redrive_execution.assert_called_once_with(
            executionArn="arn:aws:states:us-east-1:123456789012:execution:test-state-machine:random-123"
        )

    @mock.patch.object(StepFunctionHook, "conn")
    def test_redrive_execution_without_name_should_fail(self, mock_conn):
        mock_conn.redrive_execution.return_value = {"redriveDate": datetime(2024, 1, 1)}

        with pytest.raises(
            AirflowFailException, match="Execution name is required to start RedriveExecution"
        ):
            StepFunctionHook().start_execution(
                state_machine_arn="arn:aws:states:us-east-1:123456789012:stateMachine:test-state-machine",
                is_redrive_execution=True,
            )

    def test_describe_execution(self):
        hook = StepFunctionHook(aws_conn_id="aws_default", region_name="us-east-1")
        state_machine = hook.get_conn().create_state_machine(
            name="pseudo-state-machine", definition="{}", roleArn="arn:aws:iam::000000000000:role/Role"
        )

        state_machine_arn = state_machine.get("stateMachineArn")

        execution_arn = hook.start_execution(
            state_machine_arn=state_machine_arn, name=None, state_machine_input={}
        )
        response = hook.describe_execution(execution_arn)

        assert "input" in response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""This module contains the Amazon SageMaker Unified Studio Notebook operator."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.sagemaker_unified_studio import (
    SageMakerNotebookHook,
)
from airflow.providers.amazon.aws.links.sagemaker_unified_studio import (
    SageMakerUnifiedStudioLink,
)
from airflow.providers.amazon.aws.triggers.sagemaker_unified_studio import (
    SageMakerNotebookJobTrigger,
)
from airflow.providers.common.compat.sdk import AirflowException, BaseOperator, conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class SageMakerNotebookOperator(BaseOperator):
    """
    Provides Artifact execution functionality for Sagemaker Unified Studio Workflows.

    Examples:
     .. code-block:: python

        from airflow.providers.amazon.aws.operators.sagemaker_unified_studio import SageMakerNotebookOperator

        notebook_operator = SageMakerNotebookOperator(
            task_id="notebook_task",
            input_config={"input_path": "path/to/notebook.ipynb", "input_params": ""},
            output_config={"output_format": "ipynb"},
            wait_for_completion=True,
            waiter_delay=10,
            waiter_max_attempts=1440,
        )

    :param task_id: A unique, meaningful id for the task.
    :param input_config: Configuration for the input file. Input path should be specified as a relative path.
        The provided relative path will be automatically resolved to an absolute path within
        the context of the user's home directory in the IDE. Input params should be a dict.
        Example: {'input_path': 'folder/input/notebook.ipynb', 'input_params':{'key': 'value'}}
    :param output_config:  Configuration for the output format. It should include an output_format parameter to control
        the format of the notebook execution output.
        Example: {"output_formats": ["NOTEBOOK"]}
    :param compute: compute configuration to use for the notebook execution. This is a required attribute if the execution is on a remote compute.
        Example: {"instance_type": "ml.m5.large", "volume_size_in_gb": 30, "volume_kms_key_id": "", "image_details": {"ecr_uri": "string"}, "container_entrypoint": ["string"]}
    :param termination_condition: conditions to match to terminate the remote execution.
        Example: { "MaxRuntimeInSeconds": 3600 }
    :param tags: tags to be associated with the remote execution runs.
        Example: { "md_analytics": "logs" }
    :param wait_for_completion: Indicates whether to wait for the notebook execution to complete. If True, wait for completion; if False, don't wait.
    :param waiter_delay: Interval in seconds to check the notebook execution status.
    :param waiter_max_attempts: Number of attempts to wait before returning FAILED.
    :param deferrable: If True, the operator will wait asynchronously for the job to complete.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:SageMakerNotebookOperator`
    """

    operator_extra_links = (SageMakerUnifiedStudioLink(),)

    def __init__(
        self,
        task_id: str,
        input_config: dict,
        output_config: dict | None = None,
        compute: dict | None = None,
        termination_condition: dict | None = None,
        tags: dict | None = None,
        wait_for_completion: bool = True,
        waiter_delay: int = 10,
        waiter_max_attempts: int = 1440,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(task_id=task_id, **kwargs)
        self.execution_name = task_id
        self.input_config = input_config
        self.output_config = output_config or {"output_formats": ["NOTEBOOK"]}
        self.compute = compute or {}
        self.termination_condition = termination_condition or {}
        self.tags = tags or {}
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.deferrable = deferrable
        self.input_kwargs = kwargs

    @cached_property
    def notebook_execution_hook(self):
        if not self.input_config:
            raise AirflowException("input_config is required")

        if "input_path" not in self.input_config:
            raise AirflowException("input_path is a required field in the input_config")

        return SageMakerNotebookHook(
            input_config=self.input_config,
            output_config=self.output_config,
            execution_name=self.execution_name,
            compute=self.compute,
            termination_condition=self.termination_condition,
            tags=self.tags,
            waiter_delay=self.waiter_delay,
            waiter_max_attempts=self.waiter_max_attempts,
        )

    def execute(self, context: Context):
        notebook_execution = self.notebook_execution_hook.start_notebook_execution()
        execution_id = notebook_execution["execution_id"]

        if self.deferrable:
            self.defer(
                trigger=SageMakerNotebookJobTrigger(
                    execution_id=execution_id,
                    execution_name=self.execution_name,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                ),
                method_name="execute_complete",
            )
        elif self.wait_for_completion:
            response = self.notebook_execution_hook.wait_for_execution_completion(execution_id, context)
            status = response["Status"]
            log_info_message = (
                f"Notebook Execution: {self.execution_name} Status: {status}. Run Id: {execution_id}"
            )
            self.log.info(log_info_message)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from datetime import datetime
from unittest import mock as async_mock

import pytest

from airflow.providers.amazon.aws.triggers.s3 import S3KeysUnchangedTrigger, S3KeyTrigger
from airflow.triggers.base import TriggerEvent


class TestS3KeyTrigger:
    def test_serialization(self):
        """
        Asserts that the TaskStateTrigger correctly serializes its arguments
        and classpath.
        """
        trigger = S3KeyTrigger(
            bucket_key="s3://test_bucket/file",
            bucket_name="test_bucket",
            wildcard_match=True,
            metadata_keys=["Size", "LastModified"],
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.s3.S3KeyTrigger"
        assert kwargs == {
            "bucket_name": "test_bucket",
            "bucket_key": "s3://test_bucket/file",
            "wildcard_match": True,
            "aws_conn_id": "aws_default",
            "hook_params": {},
            "poke_interval": 5.0,
            "should_check_fn": False,
            "use_regex": False,
            "verify": None,
            "region_name": None,
            "botocore_config": None,
            "metadata_keys": ["Size", "LastModified"],
        }

    @pytest.mark.asyncio
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_async_conn")
    async def test_run_success(self, mock_client):
        """
        Test if the task is run is in triggerr successfully.
        """
        mock_client.return_value.return_value.check_key.return_value = True
        trigger = S3KeyTrigger(bucket_key="test_bucket/file", bucket_name="test_bucket")
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        assert task.done() is True
        result = await task
        assert result == TriggerEvent({"status": "success"})
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.check_key_async")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_async_conn")
    async def test_run_pending(self, mock_client, mock_check_key_async):
        """
        Test if the task is run is in trigger successfully and set check_key to return false.
        """
        mock_check_key_async.return_value = False
        trigger = S3KeyTrigger(bucket_key="test_bucket/file", bucket_name="test_bucket")
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        assert task.done() is False
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_files_async")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_head_object_async")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.check_key_async")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_async_conn")
    async def test_run_with_metadata(
        self,
        mock_get_async_conn,
        mock_check_key_async,
        mock_get_head_object_async,
        mock_get_files_async,
    ):
        """Test if the task retrieves metadata correctly when should_check_fn is True."""
        mock_check_key_async.return_value = True
        mock_get_files_async.return_value = ["file1.txt", "file2.txt"]

        async def fake_get_head_object_async(*args, **kwargs):
            key = kwargs.get("key")
            if key == "file1.txt":
                return {"ContentLength": 1024, "LastModified": "2023-10-01T12:00:00Z"}
            if key == "file2.txt":
                return {"ContentLength": 2048, "LastModified": "2023-10-02T12:00:00Z"}

        mock_get_head_object_async.side_effect = fake_get_head_object_async
        mock_get_async_conn.return_value.__aenter__.return_value = async_mock.AsyncMock()
        trigger = S3KeyTrigger(
            bucket_key="test_bucket/file",
            bucket_name="test_bucket",
            should_check_fn=True,
            metadata_keys=["Size", "LastModified"],
            poke_interval=0.1,  # reduce waiting time
        )
        result = await asyncio.wait_for(trigger.run().__anext__(), timeout=2)
        expected = TriggerEvent(
            {
                "status": "running",
                "files": [
                    {"Size": 1024, "LastModified": "2023-10-01T12:00:00Z", "Key": "file1.txt"},
                    {"Size": 2048, "LastModified": "2023-10-02T12:00:00Z", "Key": "file2.txt"},
                ],
            }
        )

        assert result == expected

    @pytest.mark.asyncio
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_files_async")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_head_object_async")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.check_key_async")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_async_conn")
    async def test_run_with_all_metadata(
        self, mock_get_async_conn, mock_check_key_async, mock_get_head_object_async, mock_get_files_async
    ):
        """
        Test if the task retrieves all metadata when metadata_keys contains '*'.
        """
        mock_check_key_async.return_value = True
        mock_get_files_async.return_value = ["file1.txt"]

        async def fake_get_head_object_async(*args, **kwargs):
            return {
                "ContentLength": 1024,
                "LastModified": "2023-10-01T12:00:00Z",
                "ETag": "abc123",
            }

        mock_get_head_object_async.side_effect = fake_get_head_object_async
        mock_get_async_conn.return_value.__aenter__.return_value = async_mock.AsyncMock()
        trigger = S3KeyTrigger(
            bucket_key="test_bucket/file",
            bucket_name="test_bucket",
            should_check_fn=True,
            metadata_keys=["*"],
            poke_interval=0.1,
        )
        result = await asyncio.wait_for(trigger.run().__anext__(), timeout=2)
        expected = TriggerEvent(
            {
                "status": "running",
                "files": [
                    {
                        "ContentLength": 1024,
                        "LastModified": "2023-10-01T12:00:00Z",
                        "ETag": "abc123",
                        "Key": "file1.txt",
                    }
                ],
            }
        )
        assert result == expected


class TestS3KeysUnchangedTrigger:
    def test_serialization(self):
        """
        Asserts that the S3KeysUnchangedTrigger correctly serializes its arguments
        and classpath.
        """
        trigger = S3KeysUnchangedTrigger(
            bucket_name="test_bucket",
            prefix="test",
            inactivity_period=1,
            min_objects=1,
            inactivity_seconds=0,
            previous_objects=None,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.s3.S3KeysUnchangedTrigger"
        assert kwargs == {
            "bucket_name": "test_bucket",
            "prefix": "test",
            "inactivity_period": 1,
            "min_objects": 1,
            "inactivity_seconds": 0,
            "previous_objects": set(),
            "allow_delete": True,
            "aws_conn_id": "aws_default",
            "last_activity_time": None,
            "hook_params": {},
            "verify": None,
            "region_name": None,
            "botocore_config": None,
            "polling_period_seconds": 0,
        }

    @pytest.mark.asyncio
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_async_conn")
    async def test_run_wait(self, mock_client):
        """Test if the task is run in trigger successfully."""
        mock_client.return_value.return_value.check_key.return_value = True
        trigger = S3KeysUnchangedTrigger(bucket_name="test_bucket", prefix="test")
        with mock_client:
            task = asyncio.create_task(trigger.run().__anext__())
            await asyncio.sleep(0.5)

            assert task.done() is True
            asyncio.get_event_loop().stop()

    def test_run_raise_value_error(self):
        """
        Test if the S3KeysUnchangedTrigger raises Value error for negative inactivity_period.
        """
        with pytest.raises(ValueError, match="inactivity_period must be non-negative"):
            S3KeysUnchangedTrigger(bucket_name="test_bucket", prefix="test", inactivity_period=-100)

    @pytest.mark.asyncio
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_async_conn")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.is_keys_unchanged_async")
    async def test_run_success(self, mock_is_keys_unchanged, mock_client):
        """
        Test if the task is run in triggerer successfully.
        """
        mock_is_keys_unchanged.return_value = {"status": "success"}
        trigger = S3KeysUnchangedTrigger(bucket_name="test_bucket", prefix="test")
        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "success"}) == actual

    @pytest.mark.asyncio
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.get_async_conn")
    @async_mock.patch("airflow.providers.amazon.aws.triggers.s3.S3Hook.is_keys_unchanged_async")
    async def test_run_pending(self, mock_is_keys_unchanged, mock_client):
        """Test if the task is run in triggerer successfully."""
        mock_is_keys_unchanged.return_value = {"status": "pending", "last_activity_time": datetime.now()}
        trigger = S3KeysUnchangedTrigger(bucket_name="test_bucket", prefix="test")
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)
        # TriggerEvent was not returned
        assert task.done() is False
        asyncio.get_event_loop().stop()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.hooks.elasticache_replication_group import ElastiCacheReplicationGroupHook
from airflow.providers.common.compat.sdk import AirflowException


class TestElastiCacheReplicationGroupHook:
    REPLICATION_GROUP_ID = "test-elasticache-replication-group-hook"

    REPLICATION_GROUP_CONFIG = {
        "ReplicationGroupId": REPLICATION_GROUP_ID,
        "ReplicationGroupDescription": REPLICATION_GROUP_ID,
        "AutomaticFailoverEnabled": False,
        "NumCacheClusters": 1,
        "CacheNodeType": "cache.m5.large",
        "Engine": "redis",
        "EngineVersion": "5.0.4",
        "CacheParameterGroupName": "default.redis5.0",
    }

    VALID_STATES = frozenset(
        {"creating", "available", "modifying", "deleting", "create - failed", "snapshotting"}
    )

    def setup_method(self):
        self.hook = ElastiCacheReplicationGroupHook()
        # noinspection PyPropertyAccess
        self.hook.conn = mock.Mock()

        # We need this for every test
        self.hook.conn.create_replication_group.return_value = {
            "ReplicationGroup": {"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "creating"}
        }

    def _create_replication_group(self):
        return self.hook.create_replication_group(config=self.REPLICATION_GROUP_CONFIG)

    def test_conn_not_none(self):
        assert self.hook.conn is not None

    def test_create_replication_group(self):
        response = self._create_replication_group()
        assert response["ReplicationGroup"]["ReplicationGroupId"] == self.REPLICATION_GROUP_ID
        assert response["ReplicationGroup"]["Status"] == "creating"

    def test_describe_replication_group(self):
        self._create_replication_group()

        self.hook.conn.describe_replication_groups.return_value = {
            "ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID}]
        }

        response = self.hook.describe_replication_group(replication_group_id=self.REPLICATION_GROUP_ID)
        assert response["ReplicationGroups"][0]["ReplicationGroupId"] == self.REPLICATION_GROUP_ID

    def test_get_replication_group_status(self):
        self._create_replication_group()

        self.hook.conn.describe_replication_groups.return_value = {
            "ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "available"}]
        }

        response = self.hook.get_replication_group_status(replication_group_id=self.REPLICATION_GROUP_ID)
        assert response in self.VALID_STATES

    def test_is_replication_group_available(self):
        self._create_replication_group()

        self.hook.conn.describe_replication_groups.return_value = {
            "ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "available"}]
        }

        response = self.hook.is_replication_group_available(replication_group_id=self.REPLICATION_GROUP_ID)
        assert response in (True, False)

    @mock.patch("time.sleep", return_value=None)
    def test_wait_for_availability(self, time_mock):
        self._create_replication_group()

        # Test non availability
        self.hook.conn.describe_replication_groups.return_value = {
            "ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "creating"}]
        }

        response = self.hook.wait_for_availability(
            replication_group_id=self.REPLICATION_GROUP_ID,
            max_retries=1,
            initial_sleep_time=1,  # seconds
        )
        assert response is False

        # Test availability
        self.hook.conn.describe_replication_groups.return_value = {
            "ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "available"}]
        }

        response = self.hook.wait_for_availability(
            replication_group_id=self.REPLICATION_GROUP_ID,
            max_retries=1,
            initial_sleep_time=1,  # seconds
        )
        assert response is True

    def test_delete_replication_group(self):
        self._create_replication_group()

        self.hook.conn.delete_replication_group.return_value = {
            "ReplicationGroup": {"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "deleting"}
        }

        # Wait for availability, can only delete when replication group is available
        self.hook.conn.describe_replication_groups.return_value = {
            "ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "available"}]
        }

        response = self.hook.wait_for_availability(
            replication_group_id=self.REPLICATION_GROUP_ID,
            max_retries=1,
            initial_sleep_time=1,  # seconds
        )
        assert response is True

        response = self.hook.delete_replication_group(replication_group_id=self.REPLICATION_GROUP_ID)
        assert response["ReplicationGroup"]["ReplicationGroupId"] == self.REPLICATION_GROUP_ID
        assert response["ReplicationGroup"]["Status"] == "deleting"

    def _raise_replication_group_not_found_exp(self):
        self.hook.conn.exceptions.ReplicationGroupNotFoundFault = BaseException
        return self.hook.conn.exceptions.ReplicationGroupNotFoundFault

    # noinspection PyUnusedLocal
    def _mock_describe_side_effect(self, *args, **kwargs):
        return [
            # On first call replication group is in available state, this will allow to initiate a delete
            # A replication group can only be deleted when it is in `available` state
            {"ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "available"}]},
            # On second call replication group is in deleting state
            {"ReplicationGroups": [{"ReplicationGroupId": self.REPLICATION_GROUP_ID, "Status": "deleting"}]},
            # On further calls we will assume the replication group is deleted
            self._raise_replication_group_not_found_exp(),
        ]

    @mock.patch("time.sleep", return_value=None)
    def test_wait_for_deletion(self, time_mock):
        self._create_replication_group()

        self.hook.conn.describe_replication_groups.side_effect = self._mock_describe_side_effect()

        self.hook.conn.delete_replication_group.return_value = {
            "ReplicationGroup": {"ReplicationGroupId": self.REPLICATION_GROUP_ID}
        }

        response, deleted = self.hook.wait_for_deletion(
            replication_group_id=self.REPLICATION_GROUP_ID,
            # Initial call - status is `available`
            # 1st retry - status is `deleting`
            # 2nd retry - Replication group is deleted
            max_retries=2,
            initial_sleep_time=1,
        )
        assert response["ReplicationGroup"]["ReplicationGroupId"] == self.REPLICATION_GROUP_ID
        assert deleted is True

    def test_ensure_delete_replication_group_success(self):
        self._create_replication_group()

        self.hook.conn.describe_replication_groups.side_effect = self._mock_describe_side_effect()

        self.hook.conn.delete_replication_group.return_value = {
            "ReplicationGroup": {"ReplicationGroupId": self.REPLICATION_GROUP_ID}
        }

        response = self.hook.ensure_delete_replication_group(
            replication_group_id=self.REPLICATION_GROUP_ID, initial_sleep_time=1, max_retries=2
        )

        assert response["ReplicationGroup"]["ReplicationGroupId"] == self.REPLICATION_GROUP_ID

    def test_ensure_delete_replication_group_failure(self):
        self._create_replication_group()

        self.hook.conn.describe_replication_groups.side_effect = self._mock_describe_side_effect()

        self.hook.conn.delete_replication_group.return_value = {
            "ReplicationGroup": {"ReplicationGroupId": self.REPLICATION_GROUP_ID}
        }

        with pytest.raises(AirflowException):
            # Try only 1 once with 1 sec buffer time. This will ensure that the `wait_for_deletion` loop
            # breaks quickly before the group is deleted and we get the Airflow exception
            self.hook.ensure_delete_replication_group(
                replication_group_id=self.REPLICATION_GROUP_ID, initial_sleep_time=1, max_retries=1
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import copy
from datetime import datetime
from unittest import mock

import pytest

from airflow.providers.amazon.aws.hooks.sagemaker import SageMakerHook
from airflow.providers.amazon.aws.sensors.sagemaker import SageMakerPipelineSensor
from airflow.providers.common.compat.sdk import AirflowException

DESCRIBE_PIPELINE_EXECUTION_RESPONSE = {
    "PipelineArn": "string",
    "PipelineExecutionArn": "string",
    "PipelineExecutionDisplayName": "string",
    # Status can be: "Executing" | "Stopping" | "Stopped" | "Failed" | "Succeeded"
    "PipelineExecutionStatus": "-- to be set in test --",
    "PipelineExecutionDescription": "string",
    "PipelineExperimentConfig": {"ExperimentName": "string", "TrialName": "string"},
    "FailureReason": "string",
    "CreationTime": datetime(2015, 1, 1),
    "LastModifiedTime": datetime(2015, 1, 1),
    "CreatedBy": {"UserProfileArn": "string", "UserProfileName": "string", "DomainId": "string"},
    "LastModifiedBy": {"UserProfileArn": "string", "UserProfileName": "string", "DomainId": "string"},
    "ParallelismConfiguration": {"MaxParallelExecutionSteps": 123},
    "ResponseMetadata": {
        "HTTPStatusCode": 200,
    },
}


class TestSageMakerPipelineSensor:
    @staticmethod
    def get_response_with_state(state: str):
        states = {"Executing", "Stopping", "Stopped", "Failed", "Succeeded"}
        assert state in states
        res_copy = copy.deepcopy(DESCRIBE_PIPELINE_EXECUTION_RESPONSE)
        res_copy["PipelineExecutionStatus"] = state
        return res_copy

    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(SageMakerHook, "describe_pipeline_exec")
    def test_sensor_with_failure(self, mock_describe, _):
        response_failure = self.get_response_with_state("Failed")
        mock_describe.return_value = response_failure
        sensor = SageMakerPipelineSensor(pipeline_exec_arn="ARN", task_id="test_task")

        with pytest.raises(AirflowException):
            sensor.execute(None)

        mock_describe.assert_called_once_with("ARN", True)

    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(SageMakerHook, "describe_pipeline_exec")
    def test_sensor(self, mock_describe, _):
        response_executing = self.get_response_with_state("Executing")
        response_stopping = self.get_response_with_state("Stopping")
        response_stopped = self.get_response_with_state("Stopped")
        mock_describe.side_effect = [
            response_executing,
            response_stopping,
            response_stopped,
        ]
        sensor = SageMakerPipelineSensor(pipeline_exec_arn="ARN", task_id="test_task", poke_interval=0)

        sensor.execute(None)

        assert mock_describe.call_count == 3



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.providers.amazon.aws.hooks.mwaa import MwaaHook
from airflow.providers.amazon.aws.hooks.sts import StsHook
from airflow.providers.amazon.aws.operators.mwaa import MwaaTriggerDagRunOperator
from airflow.providers.amazon.aws.sensors.mwaa import MwaaDagRunSensor, MwaaTaskSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task, timezone
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]
    from airflow.utils import timezone  # type: ignore[attr-defined,no-redef]


from system.amazon.aws.utils import SystemTestContextBuilder

DAG_ID = "example_mwaa"

# Externally fetched variables:
EXISTING_ENVIRONMENT_NAME_KEY = "ENVIRONMENT_NAME"
EXISTING_DAG_ID_KEY = "DAG_ID"
EXISTING_TASK_ID_KEY = "TASK_ID"
ROLE_WITHOUT_INVOKE_REST_API_ARN_KEY = "ROLE_WITHOUT_INVOKE_REST_API_ARN"

sys_test_context_task = (
    SystemTestContextBuilder()
    # NOTE: Creating a functional MWAA environment is time-consuming and requires
    # manually creating and configuring an S3 bucket for DAG storage and a VPC with
    # private subnets which is out of scope for this demo. To simplify this demo and
    # make it run in a reasonable time, an existing MWAA environment already
    # containing a DAG is required.
    # Here's a quick start guide to create an MWAA environment using AWS CloudFormation:
    # https://docs.aws.amazon.com/mwaa/latest/userguide/quick-start.html
    # If creating the environment using the AWS console, make sure to have a VPC with
    # at least 1 private subnet to be able to select the VPC while going through the
    # environment creation steps in the console wizard.
    # Make sure to set the environment variables with appropriate values
    .add_variable(EXISTING_ENVIRONMENT_NAME_KEY)
    .add_variable(EXISTING_DAG_ID_KEY)
    .add_variable(ROLE_WITHOUT_INVOKE_REST_API_ARN_KEY)
    .add_variable(EXISTING_TASK_ID_KEY)
    .build()
)


@task
def unpause_dag(env_name: str, dag_id: str):
    mwaa_hook = MwaaHook()
    response = mwaa_hook.invoke_rest_api(
        env_name=env_name, path=f"/dags/{dag_id}", method="PATCH", body={"is_paused": False}
    )
    return not response["RestApiResponse"]["is_paused"]


# This task in the system test verifies that the MwaaHook's IAM fallback mechanism continues to work with
# the live MWAA API. This fallback depends on parsing a specific error message from the MWAA API, so we
# want to ensure we find out if the API response format ever changes. Unit tests cover this with mocked
# responses, but this system test validates against the real API.
@task
def test_iam_fallback(role_to_assume_arn, mwaa_env_name):
    assumed_role = StsHook().conn.assume_role(
        RoleArn=role_to_assume_arn, RoleSessionName="MwaaSysTestIamFallback"
    )

    credentials = assumed_role["Credentials"]
    session = boto3.Session(
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )

    mwaa_hook = MwaaHook()
    mwaa_hook.conn = session.client("mwaa")
    response = mwaa_hook.invoke_rest_api(
        env_name=mwaa_env_name, path="/dags", method="GET", airflow_version=3
    )
    return "dags" in response["RestApiResponse"]


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_name = test_context[EXISTING_ENVIRONMENT_NAME_KEY]
    trigger_dag_id = test_context[EXISTING_DAG_ID_KEY]
    task_id = test_context[EXISTING_TASK_ID_KEY]
    restricted_role_arn = test_context[ROLE_WITHOUT_INVOKE_REST_API_ARN_KEY]

    # [START howto_operator_mwaa_trigger_dag_run]
    trigger_dag_run = MwaaTriggerDagRunOperator(
        task_id="trigger_dag_run",
        env_name=env_name,
        logical_date=datetime.now(timezone.utc).isoformat(),
        trigger_dag_id=trigger_dag_id,
        wait_for_completion=True,
        airflow_version=3,
    )
    # [END howto_operator_mwaa_trigger_dag_run]

    # [START howto_sensor_mwaa_task]
    wait_for_task = MwaaTaskSensor(
        task_id="wait_for_task",
        external_env_name=env_name,
        external_dag_id=trigger_dag_id,
        external_task_id=task_id,
        poke_interval=5,
    )
    # [END howto_sensor_mwaa_task]

    # [START howto_sensor_mwaa_dag_run]
    wait_for_dag_run = MwaaDagRunSensor(
        task_id="wait_for_dag_run",
        external_env_name=env_name,
        external_dag_id=trigger_dag_id,
        external_dag_run_id="{{ task_instance.xcom_pull(task_ids='trigger_dag_run')['RestApiResponse']['dag_run_id'] }}",
        poke_interval=5,
    )
    # [END howto_sensor_mwaa_dag_run]

    trigger_dag_run_dont_wait = MwaaTriggerDagRunOperator(
        task_id="trigger_dag_run_dont_wait",
        env_name=env_name,
        logical_date=datetime.now(timezone.utc).isoformat(),
        trigger_dag_id=trigger_dag_id,
        wait_for_completion=False,
        airflow_version=3,
    )

    wait_for_task_concurrent = MwaaTaskSensor(
        task_id="wait_for_task_concurrent",
        external_env_name=env_name,
        external_dag_id=trigger_dag_id,
        external_task_id=task_id,
        poke_interval=5,
    )

    test_context >> [
        unpause_dag(env_name, trigger_dag_id),
        test_iam_fallback(restricted_role_arn, env_name),
        trigger_dag_run,
        trigger_dag_run_dont_wait,
    ]
    chain(trigger_dag_run, wait_for_task, wait_for_dag_run)
    chain(trigger_dag_run_dont_wait, wait_for_task_concurrent)

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime as dt
import json
import random
from datetime import datetime

import boto3
from sqlalchemy import select

from airflow import settings
from airflow.models import Connection
from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.operators.kinesis_analytics import (
    KinesisAnalyticsV2CreateApplicationOperator,
    KinesisAnalyticsV2StartApplicationOperator,
    KinesisAnalyticsV2StopApplicationOperator,
)
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.kinesis_analytics import (
    KinesisAnalyticsV2StartApplicationCompletedSensor,
    KinesisAnalyticsV2StopApplicationCompletedSensor,
)
from airflow.providers.amazon.aws.transfers.http_to_s3 import HttpToS3Operator

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task, task_group
else:
    # Airflow 2 path
    from airflow.decorators import task, task_group  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import SystemTestContextBuilder

ROLE_ARN_KEY = "ROLE_ARN"
sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

DAG_ID = "example_kinesis_analytics"


@task_group
def kinesis_analytics_v2_workflow():
    # [START howto_operator_create_application]
    create_application = KinesisAnalyticsV2CreateApplicationOperator(
        task_id="create_application",
        application_name=application_name,
        runtime_environment="FLINK-1_18",
        service_execution_role=test_context[ROLE_ARN_KEY],
        create_application_kwargs={
            "ApplicationConfiguration": {
                "FlinkApplicationConfiguration": {
                    "ParallelismConfiguration": {
                        "ConfigurationType": "CUSTOM",
                        "Parallelism": 2,
                        "ParallelismPerKPU": 1,
                        "AutoScalingEnabled": False,
                    }
                },
                "EnvironmentProperties": {
                    "PropertyGroups": [
                        {
                            "PropertyGroupId": "BlueprintMetadata",
                            "PropertyMap": {
                                "AWSRegion": region_name,
                                "BlueprintName": "KDS_FLINK-DATASTREAM-JAVA_S3",
                                "BucketName": f"s3://{bucket_name}/",
                                "PartitionFormat": "yyyy-MM-dd-HH",
                                "StreamInitialPosition": "TRIM_HORIZON",
                                "StreamName": stream_name,
                            },
                        },
                    ]
                },
                "ApplicationCodeConfiguration": {
                    "CodeContent": {
                        "S3ContentLocation": {
                            "BucketARN": f"arn:aws:s3:::{bucket_name}",
                            "FileKey": "code/kds-to-s3-datastream-java-1.0.1.jar",
                        },
                    },
                    "CodeContentType": "ZIPFILE",
                },
            }
        },
    )
    # [END howto_operator_create_application]

    # [START howto_operator_start_application]
    start_application = KinesisAnalyticsV2StartApplicationOperator(
        task_id="start_application",
        application_name=application_name,
    )
    # [END howto_operator_start_application]
    start_application.wait_for_completion = False

    # [START howto_sensor_start_application]
    await_start_application = KinesisAnalyticsV2StartApplicationCompletedSensor(
        task_id="await_start_application",
        application_name=application_name,
    )
    # [END howto_sensor_start_application]

    # [START howto_operator_stop_application]
    stop_application = KinesisAnalyticsV2StopApplicationOperator(
        task_id="stop_application",
        application_name=application_name,
    )
    # [END howto_operator_stop_application]
    stop_application.wait_for_completion = False
    # With the default `force=False` the test fails  because the data retention policy is blocking the deletion.
    stop_application.force = True

    # [START howto_sensor_stop_application]
    await_stop_application = KinesisAnalyticsV2StopApplicationCompletedSensor(
        task_id="await_stop_application",
        application_name=application_name,
    )
    # [END howto_sensor_stop_application]

    @task(trigger_rule=TriggerRule.ALL_DONE)
    def delete_application(app_name: str):
        kinesis_analytics_v2_hook = KinesisAnalyticsV2Hook()
        response = kinesis_analytics_v2_hook.conn.describe_application(ApplicationName=app_name)
        kinesis_analytics_v2_hook.conn.delete_application(
            ApplicationName=app_name, CreateTimestamp=response["ApplicationDetail"]["CreateTimestamp"]
        )

    chain(
        create_application,
        start_application,
        await_start_application,
        stop_application,
        await_stop_application,
        delete_application(application_name),
    )


@task_group
def copy_jar_to_s3(bucket: str):
    """

    Copy application code to S3 using HttpToS3Operator.

    :param bucket: Name of the Amazon S3 bucket.
    """

    @task
    def create_connection(conn_id):
        conn = Connection(
            conn_id=conn_id,
            conn_type="http",
            host="https://github.com/",
        )
        if settings.Session is None:
            raise RuntimeError("Session not configured. Call configure_orm() first.")
        session = settings.Session()
        session.add(conn)
        session.commit()

    @task(trigger_rule=TriggerRule.ALL_DONE)
    def delete_connection(conn_id: str):
        if settings.Session is None:
            raise RuntimeError("Session not configured. Call configure_orm() first.")
        session = settings.Session()
        conn_to_details = session.scalar(select(Connection).where(Connection.conn_id == conn_id))
        session.delete(conn_to_details)
        session.commit()

    copy_jar_file = HttpToS3Operator(
        task_id="copy_jar_file",
        http_conn_id=http_conn_id,
        endpoint="awslabs/managed-service-for-apache-flink-blueprints/releases/download/v2.0.1/kds-to-s3-datastream-java-1.0.1.jar",
        s3_bucket=bucket,
        s3_key="code/kds-to-s3-datastream-java-1.0.1.jar",
    )

    chain(create_connection(http_conn_id), copy_jar_file, delete_connection(http_conn_id))


@task
def create_kinesis_stream(stream: str, region: str):
    """
    Create kinesis stream and put some sample data.

    :param stream: Name of the kinesis stream.
    :param region: Region name
    """
    client = boto3.client("kinesis", region_name=region)
    client.create_stream(StreamName=stream, ShardCount=1, StreamModeDetails={"StreamMode": "PROVISIONED"})
    account_id = boto3.client("sts").get_caller_identity()["Account"]
    waiter = client.get_waiter("stream_exists")
    waiter.wait(StreamName=stream, WaiterConfig={"Delay": 60, "MaxAttempts": 4})

    def get_data():
        return {
            "event_time": dt.datetime.now().isoformat(),
            "ticker": random.choice(["AAPL", "AMZN", "MSFT", "INTC", "TBV"]),
            "price": round(random.random() * 100, 2),
        }

    for _ in range(2):
        data = get_data()
        client.put_record(
            StreamARN=f"arn:aws:kinesis:{region}:{account_id}:stream/{stream}",
            Data=json.dumps(data),
            PartitionKey=data["ticker"],
        )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_kinesis_stream(stream: str, region: str):
    client = boto3.client("kinesis", region_name=region)
    client.delete_stream(StreamName=stream, EnforceConsumerDeletion=True)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]
    bucket_name = f"{env_id}-kinesis-analytics"
    application_name = f"{env_id}-test-app"
    http_conn_id = f"{env_id}-git"
    region_name = boto3.session.Session().region_name
    stream_name = f"{env_id}-test-stream"

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        create_kinesis_stream(stream=stream_name, region=region_name),
        copy_jar_to_s3(bucket=bucket_name),
        # TEST BODY
        kinesis_analytics_v2_workflow(),
        # TEST TEARDOWN
        delete_kinesis_stream(stream=stream_name, region=region_name),
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import boto3
import pytest
from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.hooks.batch_client import BatchClientHook


class TestCustomBatchServiceWaiters:
    JOB_ID = "test_job_id"

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self, monkeypatch):
        self.client = boto3.client("batch", region_name="eu-west-3")
        monkeypatch.setattr(BatchClientHook, "conn", self.client)

    @pytest.fixture
    def mock_describe_jobs(self):
        """Mock ``BatchClientHook.Client.describe_jobs`` method."""
        with mock.patch.object(self.client, "describe_jobs") as m:
            yield m

    def test_service_waiters(self):
        hook_waiters = BatchClientHook(aws_conn_id=None).list_waiters()
        assert "batch_job_complete" in hook_waiters

    @staticmethod
    def describe_jobs(status: str):
        """
        Helper function for generate minimal DescribeJobs response for a single job.
        https://docs.aws.amazon.com/batch/latest/APIReference/API_DescribeJobs.html
        """
        return {
            "jobs": [
                {
                    "status": status,
                },
            ],
        }

    def test_job_succeeded(self, mock_describe_jobs):
        """Test job succeeded"""
        mock_describe_jobs.side_effect = [
            self.describe_jobs(BatchClientHook.RUNNING_STATE),
            self.describe_jobs(BatchClientHook.SUCCESS_STATE),
        ]
        waiter = BatchClientHook(aws_conn_id=None).get_waiter("batch_job_complete")
        waiter.wait(jobs=[self.JOB_ID], WaiterConfig={"Delay": 0.01, "MaxAttempts": 2})

    def test_job_failed(self, mock_describe_jobs):
        """Test job failed"""
        mock_describe_jobs.side_effect = [
            self.describe_jobs(BatchClientHook.RUNNING_STATE),
            self.describe_jobs(BatchClientHook.FAILURE_STATE),
        ]
        waiter = BatchClientHook(aws_conn_id=None).get_waiter("batch_job_complete")

        with pytest.raises(WaiterError, match="Waiter encountered a terminal failure state"):
            waiter.wait(jobs=[self.JOB_ID], WaiterConfig={"Delay": 0.01, "MaxAttempts": 2})
